from qtpy.QtWidgets import (
    QMainWindow,
    QLabel,
    QVBoxLayout,
    QPushButton,
    QWidget,
    QListWidget,
    QHBoxLayout,
    QLineEdit,
    QTreeView,
    QSplitter,
)
from qtpy.QtCore import Qt
from qtpy import QtGui
from qtpy.QtGui import QStandardItemModel
from pkg_resources import resource_filename

from brainrender.gui.style import style, tree_css, update_css, _themes
from brainrender.gui.widgets.tree import StandardItem


class UI(QMainWindow):
    buttons = {}

    right_navbar_button_names = [
        "add brain regions",
        "add cells",
        "add from file",
    ]

    central_column_button_names = [
        "take screenshot",
        "reset",
        "top",
        "side1",
        "side2",
        "front",
    ]

    def __init__(self, theme="dark", **kwargs):
        super().__init__()

        # Get palette
        self.palette = _themes[theme]
        self.theme = theme

        # set the title and icon of main window
        self.setWindowTitle("BRAINGLOBE - brainrender GUI")

        logo_path = resource_filename(
            "brainrender.gui.icons", "BG_logo_mini.svg"
        )
        self.setWindowIcon(QtGui.QIcon(logo_path))

        # set the size of window
        self.Width = 3000
        self.height = int(0.618 * self.Width)
        self.resize(self.Width, self.height)

        # Create UI
        self.get_icons()
        self.initUI()
        self.setStyleSheet(update_css(style, self.palette))

    def get_icons(self):
        """
        Gets the correct path to the icons
        depending on the theme chosen
        """
        self.palette["branch_closed_img"] = resource_filename(
            "brainrender.gui.icons", f"right_{self.theme}.svg"
        )
        self.palette["branch_opened_img"] = resource_filename(
            "brainrender.gui.icons", f"down_{self.theme}.svg"
        )
        self.palette["checked_img"] = resource_filename(
            "brainrender.gui.icons", f"checkedbox_{self.theme}.svg"
        )
        self.palette["unchecked_img"] = resource_filename(
            "brainrender.gui.icons", f"box_{self.theme}.svg"
        )

    def make_left_navbar(self):
        """
        Creates the structures tree hierarchy widget and populates
        it with structures names from the brainglobe-api's Atlas.hierarchy
        tree view.
        """
        # Create QTree widget
        treeView = QTreeView()
        treeView.setExpandsOnDoubleClick(False)
        treeView.setHeaderHidden(True)
        treeView.setStyleSheet(update_css(tree_css, self.palette))
        treeView.setWordWrap(False)

        treeModel = QStandardItemModel()
        rootNode = treeModel.invisibleRootItem()

        # Add element's hierarchy
        tree = self.scene.atlas.hierarchy
        items = {}
        for n, node in enumerate(tree.expand_tree()):
            # Get Node info
            node = tree.get_node(node)
            if node.tag in ["VS", "fiber tracts"]:
                continue

            # Get brainregion name
            name = self.scene.atlas._get_from_structure(
                node.identifier, "name"
            )

            # Create Item
            item = StandardItem(
                name,
                node.tag,
                tree.depth(node.identifier),
                self.palette["text"],
            )

            # Get/assign parents
            parent = tree.parent(node.identifier)
            if parent is not None:
                if parent.identifier not in items.keys():
                    continue
                else:
                    items[parent.identifier].appendRow(item)

            # Keep track of added nodes
            items[node.identifier] = item
            if n == 0:
                root = item

        # Finish up
        rootNode.appendRow(root)
        treeView.setModel(treeModel)
        treeView.expandToDepth(2)
        self.treeView = treeView

        return treeView

    def make_right_navbar(self):
        """
        Creates the widgets in the right navbar.
        """
        # make layout
        layout = QVBoxLayout()

        # Add label
        layout.addWidget(QLabel("Add actors"))

        # Add buttons
        for bname in self.right_navbar_button_names:
            btn = QPushButton(bname.capitalize(), self)
            btn.setObjectName(bname.replace(" ", "_"))
            self.buttons[bname.replace(" ", "_")] = btn
            layout.addWidget(btn)

        # Add label
        lbl = QLabel("Actors")
        lbl.setObjectName("LabelWithBorder")
        layout.addWidget(lbl)

        # add list widget
        self.actors_list = QListWidget()
        self.actors_list.setObjectName("actors_list")
        layout.addWidget(self.actors_list)

        # Add label
        lbl = QLabel("Actor properties")
        lbl.setObjectName("LabelWithBorder")
        layout.addWidget(lbl)

        # Actor Alpha
        alphalabel = QLabel("Alpha")
        alphalabel.setObjectName("PropertyName")
        self.alpha_textbox = QLineEdit(self)
        self.alpha_textbox.setObjectName("Property")
        layout.addWidget(alphalabel)
        layout.addWidget(self.alpha_textbox)

        # Actor Color
        colorlabel = QLabel("Color")
        colorlabel.setObjectName("PropertyName")
        self.color_textbox = QLineEdit(self)
        self.color_textbox.setObjectName("Property")
        layout.addWidget(colorlabel)
        layout.addWidget(self.color_textbox)

        # Add label
        lbl = QLabel("Show structures tree")
        lbl.setObjectName("LabelWithBorder")
        layout.addWidget(lbl)

        btn = QPushButton("Show structures tree", self)
        self.buttons[btn.text().lower().replace(" ", "_")] = btn
        layout.addWidget(btn)

        # set spacing
        layout.addStretch(5)
        layout.setSpacing(20)

        # make widget
        widget = QWidget()
        widget.setObjectName("RightNavbar")
        widget.setLayout(layout)

        return widget

    def make_central_column(self):
        """
        Creates vtkWidget for the vedo plotter and a few
        useful buttons, for the central part of the GUI
        """
        # Create layout, add canvas and buttons
        layout = QVBoxLayout()
        layout.addWidget(self.vtkWidget)

        # Make buttons
        boxes = [QHBoxLayout(), QHBoxLayout()]
        for n, bname in enumerate(self.central_column_button_names):
            btn = QPushButton(bname.capitalize(), self)
            btn.setObjectName(bname.replace(" ", "_"))
            self.buttons[bname.replace(" ", "_")] = btn

            if n == 0:
                boxes[0].addWidget(btn)
            else:
                boxes[1].addWidget(btn)

        hlayout = QHBoxLayout()

        widget = QWidget()
        widget.setLayout(boxes[0])
        widget.setObjectName("ScreenshotButtonLayout")
        hlayout.addWidget(widget)

        widget = QWidget()
        widget.setLayout(boxes[1])
        hlayout.addWidget(widget)

        widget = QWidget()
        widget.setObjectName("CentralColumn_buttons")
        widget.setLayout(hlayout)

        layout.addWidget(widget)

        # make widget
        widget = QWidget()
        widget.setObjectName("CentralColumn")
        widget.setLayout(layout)

        return widget

    def initUI(self):
        """
        Define UI elements of the app's main window
        """
        # Create navbars
        self.treeView = self.make_left_navbar()

        # Make overall layout
        main_layout = QHBoxLayout()

        splitter = QSplitter(Qt.Horizontal)
        splitter.addWidget(self.treeView)
        splitter.addWidget(self.make_central_column())
        splitter.addWidget(self.make_right_navbar())

        splitter.setSizes([200, 700, 10])
        main_layout.addWidget(splitter)

        self.treeView.setHidden(True)

        # Create main window widget
        main_widget = QWidget()
        main_widget.setObjectName("MainWidget")
        main_widget.setLayout(main_layout)
        self.setCentralWidget(main_widget)



import vtk
from loguru import logger

from brainrender.cameras import cameras


def get_camera(camera):
    """
    Returns the parameters for a pre-defined camera

    :param camera: str
    """
    return cameras[camera]


def check_camera_param(camera):
    """
    Check that a dictionary of camera parameters
    is complete. Must have entries:
    ["position", "focal", "viewup", "distance", "clipping"]

    :param camera: str, dict
    """
    if isinstance(camera, str):
        return get_camera(camera)
    else:
        params = ["pos", "viewup", "clippingRange"]
        for param in params:
            if param not in list(camera.keys()):
                raise ValueError(
                    f"Camera parameters dict should include the following keys: {params}, missing: {param}"
                )
        if "focalPoint" not in camera.keys():
            camera["focalPoint"] = None
        return camera


def set_camera_params(camera, params):
    """
    Set camera parameters
    :param camera: camera obj
    :param params: dictionary of camera parameters
    """
    logger.debug(f"Setting camera parameters: {params}")
    # Apply camera parameters
    camera.SetPosition(params["pos"])
    camera.SetViewUp(params["viewup"])
    camera.SetClippingRange(params["clippingRange"])

    if "focalPoint" in params.keys() and params["focalPoint"] is not None:
        camera.SetFocalPoint(params["focalPoint"])
    if "distance" in params.keys():
        camera.SetDistance(params["distance"])


def set_camera(scene, camera):
    """
    Sets the position of the camera of a brainrender scene.

    :param scene: instance of Scene
    :param camera: either a string with the name of one of the pre-defined cameras, or
                    a dictionary of camera parameters.

    """
    if camera is None:
        return

    if not isinstance(camera, vtk.vtkCamera):
        # Get camera params
        camera = check_camera_param(camera)

        # set params
        try:
            set_camera_params(scene.plotter.camera, camera)
        except AttributeError:
            return None
    else:
        scene.plotter.camera = camera
    return camera


def get_camera_params(scene=None, camera=None):
    """
    Given an active brainrender scene or a camera, it return
    the camera parameters.

    :param scene: instance of Scene whose camera is to be used
    :param camera: camera obj
    """

    def clean(val):
        if isinstance(val, tuple):
            return tuple((round(v) for v in val))
        else:
            return round(val)

    if scene is not None:
        if not scene.is_rendered:
            scene.render(interactive=False)
        cam = scene.plotter.camera
    else:
        cam = camera

    params = dict(
        pos=clean(cam.GetPosition()),
        focalPoint=clean(cam.GetFocalPoint()),
        viewup=clean(cam.GetViewUp()),
        distance=clean(cam.GetDistance()),
        clippingRange=clean(cam.GetClippingRange()),
        # orientation=clean(cam.GetOrientation()),
    )
    return params



from vedo import shapes
from vedo import Mesh
from loguru import logger

from brainrender.actor import Actor


class Cylinder(Actor):
    def __init__(self, pos, root, color="powderblue", alpha=1, radius=350):
        """
        Cylinder class creates a cylinder mesh between a given
        point and the brain's surface.

        :param pos: list, np.array of ap, dv, ml coordinates.
            If an actor is passed, get's the center of mass instead
        :param root: brain root Actor or mesh object
        :param color: str, color
        :param alpha: float
        :param radius: float
        """

        # Get pos
        if isinstance(pos, Mesh):
            pos = pos.points().mean(axis=0)
        elif isinstance(pos, Actor):
            pos = pos.center
        logger.debug(f"Creating Cylinder actor at: {pos}")

        # Get point at top of cylinder
        top = pos.copy()
        top[1] = root.bounds()[2] - 500

        # Create mesh and Actor
        mesh = shapes.Cylinder(pos=[top, pos], c=color, r=radius, alpha=alpha)
        Actor.__init__(self, mesh, name="Cylinder", br_class="Cylinder")



import os
from loguru import logger
import sys
from pathlib import Path
from rich.logging import RichHandler

try:
    from pyinspect import install_traceback

    install_traceback()
except ImportError:
    pass  # fails in notebooks

from brainrender import settings
from brainrender.scene import Scene
import brainrender.actors
from brainrender.video import VideoMaker, Animation

base_dir = Path(os.path.join(os.path.expanduser("~"), ".brainrender"))
base_dir.mkdir(exist_ok=True)

__version__ = "2.0.5.0"

# set logger level
def set_logging(level="INFO", path=None):
    """
    Sets loguru to save all logs to a file i
    brainrender's base directory and to print
    to stdout only logs >= to a given level
    """
    logger.remove()
    # logger.add(sys.stdout, level=level)

    path = path or str(base_dir / "log.log")
    if Path(path).exists():
        Path(path).unlink()
    logger.add(path, level="DEBUG")

    if level == "DEBUG":
        logger.configure(
            handlers=[
                {
                    "sink": RichHandler(level="WARNING", markup=True),
                    "format": "{message}",
                }
            ]
        )


if not settings.DEBUG:
    set_logging()
else:
    set_logging(level="DEBUG")



from qtpy.QtWidgets import QApplication
import sys
import click

from brainrender.gui.app import App


def launch(*args, atlas_name=None, output=None, screenshots_folder=None):
    """
    Launches the application
    """
    if output is None:
        screenshot_kwargs = output
    else:
        screenshot_kwargs = screenshots_folder

    app = QApplication(sys.argv)
    app.setApplicationName("Brainrender GUIs")
    ex = App(
        *args, atlas_name=atlas_name, screenshots_folder=screenshots_folder
    )
    app.aboutToQuit.connect(ex.onClose)
    ex.show()
    sys.exit(app.exec_())


@click.command()
@click.option("-x", "--axes", is_flag=True, default=False)
@click.option("-a", "--atlas", default=None)
@click.option("-o", "--output", default=None)
def clilaunch(atlas=None, axes=False, output=None):
    app = QApplication(sys.argv)
    app.setApplicationName("Brainrender GUIs")
    ex = App(atlas_name=atlas, axes=axes, screenshots_folder=output)
    app.aboutToQuit.connect(ex.onClose)
    ex.show()
    sys.exit(app.exec_())



_themes = {
    "dark": {
        "folder": "dark",
        "background": "rgb(38, 41, 48)",
        "foreground": "rgb(65, 72, 81)",
        "primary": "rgb(90, 98, 108)",
        "secondary": "rgb(134, 142, 147)",
        "highlight": "rgb(106, 115, 128)",
        "text": "rgb(240, 241, 242)",
        "icon": "rgb(209, 210, 212)",
        "warning": "rgb(153, 18, 31)",
        "current": "rgb(0, 122, 204)",
        "syntax_style": "native",
        "console": "rgb(0, 0, 0)",
        "canvas": "black",
    },
    "light": {
        "folder": "light",
        "background": "rgb(239, 235, 233)",
        "foreground": "rgb(214, 208, 206)",
        "primary": "rgb(188, 184, 181)",
        "secondary": "rgb(150, 146, 144)",
        "highlight": "rgb(163, 158, 156)",
        "text": "rgb(59, 58, 57)",
        "icon": "rgb(107, 105, 103)",
        "warning": "rgb(255, 18, 31)",
        "current": "rgb(253, 240, 148)",
        "syntax_style": "default",
        "console": "rgb(255, 255, 255)",
        "canvas": "white",
    },
}

palette = {
    "folder": "dark",
    "background": "rgb(38, 41, 48)",
    "foreground": "rgb(65, 72, 81)",
    "primary": "rgb(90, 98, 108)",
    "secondary": "rgb(134, 142, 147)",
    "highlight": "rgb(106, 115, 128)",
    "text": "rgb(240, 241, 242)",
    "icon": "rgb(209, 210, 212)",
    "warning": "rgb(153, 18, 31)",
    "current": "rgb(0, 122, 204)",
    "syntax_style": "native",
    "console": "rgb(0, 0, 0)",
    "canvas": "black",
}

style = """
QWidget#LeftNavbar {
    background-color: BGCOLOR;
    border: 2px solid TXTCOLOR;
    border-radius: 24px;
    margin: 12px 12px;
    padding: 24px 12px;
}
QWidget#RightNavbar {
    background-color: BGCOLOR;
    margin: 12px 12px;
    padding: 24px 12px;
}
QWidget#MainWidget {
    background-color: BGCOLOR;
    padding: 48px;
}

QWidget#ScreenshotButtonLayout{
    border-right: 2px solid TXTCOLOR;
    max-width: 350px;
}


QVTKRenderWindowInteractor{
    border-radius: 12px;
    width: 200px;
}

QPushButton { 
    background-color: FGCOLOR;
    color: TXTCOLOR;
    border-radius: 8px;
    padding: 6px;
    font-size: 14pt;
    margin: 4px 34px;
    min-width: 50px;

}
QPushButton:hover {
    border: 1px solid TXTCOLOR;
}


QLabel {
    color: TXTCOLOR;
    font-size: 16pt;
    font-weight: 700;
    margin: 12px 24px;
}
QLabel#PopupLabel {
    color: TXTCOLOR;
    font-size: 14pt;
    font-weight: 400;
    margin: 12px;
}
QLabel#PropertyName {
    font-size: 14pt;
    font-weight: 500;
    margin: 0px 24px;
}
QLabel#LabelWithBorder {
    border-top: 1px solid TXTCOLOR;
    padding-top: 12px;
}

QListWidget#actors_list {
    background-color: FGCOLOR;
    color: TXTCOLOR;
    border-radius: 8px;
    padding: 6px;
    font-size: 14pt;
    margin: 4px 34px;
}

QLineEdit {
    background-color: FGCOLOR;
    color: TXTCOLOR;
    border-radius: 8px;
    padding: 6px;
    font-size: 12pt;
    height: 48px;
    min-width: 600px;
    margin: 4px 34px;
    width: 80%
}
QLineEdit#Property {
    min-width: none;
    height: 32px;
    padding: 4px 6px;
}
QDialog {
    background-color: BGCOLOR;
}
QPushButton#RegionsButton {
    width: 60%;
}

"""


# for ref: https://doc.qt.io/qt-5/stylesheet-examples.html#customizing-qtreeview
tree_css = """
QTreeView {
    background-color: BGCOLOR; 
    border-radius: 12px; 
    padding: 20px 12px;
} 
QTreeView::indicator:checked {
    image: url(CHECKED_IMG);
}
QTreeView::indicator:unchecked {
    image: url(UNCHECKED_IMG);
} 


QTreeView::branch:has-children:!has-siblings:closed,QTreeView::branch:closed:has-children:has-siblings {
    border-image: none;
    image: url(CLOSED_IMG);
}
QTreeView::branch:open:has-children:!has-siblings,QTreeView::branch:open:has-children:has-siblings  {
    border-image: none;
    image: url(OPENED_IMG);
}

"""


def update_css(css, palette):
    """
    Updates a CSS string with values
    from the palette chosen.
    """

    def path(raw_path):
        return raw_path.replace("\\", "/")

    css = css.replace("FGCOLOR", palette["foreground"])
    css = css.replace("BGCOLOR", palette["background"])
    css = css.replace("TXTCOLOR", palette["text"])
    css = css.replace("HIGHLIGHT", palette["highlight"])

    css = css.replace("CLOSED_IMG", path(palette["branch_closed_img"]))
    css = css.replace("OPENED_IMG", path(palette["branch_opened_img"]))

    css = css.replace("UNCHECKED_IMG", path(palette["unchecked_img"]))
    css = css.replace("CHECKED_IMG", path(palette["checked_img"]))

    return css



from vedo import Volume as VedoVolume
import numpy as np

from pathlib import Path
from loguru import logger

from brainrender.actor import Actor


class Volume(Actor):
    def __init__(
        self,
        griddata,
        voxel_size=1,
        cmap="bwr",
        min_quantile=None,
        min_value=None,
        name=None,
        br_class=None,
        as_surface=True,
        **volume_kwargs,
    ):
        """
        Takes a 3d numpy array with volumetric data
        and returns an Actor with mesh: vedo.Volume.isosurface or a vedo.Volume.
        BY default the volume is represented as a surface

        To extract the surface:
            The isosurface needs a lower bound threshold, this can be
            either a user defined hard value (min_value) or the value
            corresponding to some percentile of the grid data.

        :param griddata: np.ndarray, 3d array with grid data. Can also be a vedo Volume
            or a file path pointing to a .npy file
        :param griddata: np.ndarray, 3d array with grid data
        :param voxel_size: int, size of each voxel in microns
        :param min_quantile: float, percentile for threshold
        :param min_value: float, value for threshold
        :param cmap: str, name of colormap to use
        :param as_surface, bool. default True. If True
            a surface mesh is returned instead of the whole volume
        :param volume_kwargs: keyword arguments for vedo's Volume class
        """
        logger.debug(f"Creating a Volume actor")
        # Create mesh
        color = volume_kwargs.pop("c", "viridis")
        if isinstance(griddata, np.ndarray):
            # create volume from data
            mesh = self._from_numpy(griddata, voxel_size, color, **volume_kwargs)
        elif isinstance(griddata, (str, Path)):
            # create from .npy file
            mesh = self._from_file(griddata, voxel_size, color, **volume_kwargs)
        else:
            mesh = griddata  # assume a vedo Volume was passed

        if as_surface:
            # Get threshold
            if min_quantile is None and min_value is None:
                th = 0
            elif min_value is not None:
                th = min_value
            else:
                th = np.percentile(griddata.ravel(), min_quantile)

            mesh = mesh.legosurface(vmin=th, cmap=cmap)

        Actor.__init__(
            self, mesh, name=name or "Volume", br_class=br_class or "Volume"
        )

    def _from_numpy(self, griddata, voxel_size, color, **volume_kwargs):
        ''' 
            Creates a vedo.Volume actor from a 3D numpy array
            with volume data
        '''
        return VedoVolume(
                griddata,
                spacing=[voxel_size, voxel_size, voxel_size],
                c=color,
                **volume_kwargs,
            )

    def _from_file(self, filepath, voxel_size, color, **volume_kwargs):
        ''' 
            Loads a .npy file and returns a vedo Volume actor.
        '''
        filepath = Path(filepath)
        if not filepath.exists():
            raise FileExistsError(f'Loading volume from file, file not found: {filepath}')
        if not filepath.suffix == '.npy':
            raise ValueError('Loading volume from file only accepts .npy files')

        return self._from_numpy(
            np.load(str(filepath)),
            voxel_size, color, **volume_kwargs
        )



import pandas as pd
import os
import sys
from loguru import logger


from brainrender.atlas_specific.allen_brain_atlas.gene_expression.ge_utils import (
    check_gene_cached,
    load_cached_gene,
    download_and_cache,
)

from brainrender import base_dir
from brainrender._io import request, fail_on_no_connection
from brainrender.actors import Volume


class GeneExpressionAPI:
    voxel_size = 200  # um
    grid_size = [58, 41, 67]  # number of voxels along each direction

    all_genes_url = (
        "http://api.brain-map.org/api/v2/data/query.json?criteria="
        + "model::Gene,"
        + "rma::criteria,products[abbreviation$eq'DevMouse'],"
        + "rma::options,[tabular$eq'genes.id','genes.acronym+as+gene_symbol','genes.name+as+gene_name',"
        + "'genes.entrez_id+as+entrez_gene_id','genes.homologene_id+as+homologene_group_id'],"
        + "[order$eq'genes.acronym']"
        + "&num_rows=all&start_row=0"
    )

    gene_experiments_url = (
        "http://api.brain-map.org/api/v2/data/query.json?criteria=model::SectionDataSet,"
        + "rma::criteria,[failed$eq'false'],products[abbreviation$eq'Mouse'],genes[acronym$eq'-GENE_SYMBOL-']"
    )

    download_url = "http://api.brain-map.org/grid_data/download/EXP_ID?include=energy,intensity,density"

    gene_expression_cache = base_dir / "GeneExpressionCache"
    gene_name = None

    def __init__(self):
        # Get metadata about all available genes
        self.genes = None  # when necessary gene data can be downloaded with self.get_all_genes
        self.gene_expression_cache.mkdir(exist_ok=True)

    @fail_on_no_connection
    def get_all_genes(self):
        """
        Download metadata about all the genes available in the Allen gene expression dataset
        """
        res = request(self.all_genes_url)
        return pd.DataFrame(res.json()["msg"])

    def get_gene_id_by_name(self, gene_name):
        self.gene_name = self.gene_name or gene_name
        if self.genes is None:
            self.genes = self.get_all_genes()

        if gene_name not in self.genes.gene_symbol.values:
            print(
                f"Gene name {gene_name} doesnt appear in the genes dataset, nothing to return\n"
                + "You can search for you gene here: https://mouse.brain-map.org/"
            )
            return None
        else:
            return int(
                self.genes.loc[self.genes.gene_symbol == gene_name].id.values[
                    0
                ]
            )

    def get_gene_symbol_by_id(self, gene_id):
        if self.genes is None:
            self.genes = self.get_all_genes()

        return self.genes.loc[
            self.genes.id == str(gene_id)
        ].gene_symbol.values[0]

    @fail_on_no_connection
    def get_gene_experiments(self, gene):
        """
        Given a gene_symbol it returns the list of ISH
        experiments for this gene

        :param gene_symbol: str
        """
        url = self.gene_experiments_url.replace("-GENE_SYMBOL-", gene)
        data = request(url).json()["msg"]

        if not len(data):
            print(f"No experiment found for gene {gene}")
            return None
        else:
            return [d["id"] for d in data]

    @fail_on_no_connection
    def download_gene_data(self, gene):
        """
        Downloads a gene's data from the Allen Institute
        Gene Expression dataset and saves to cache.
        See: http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data

        :param gene: int, the gene_id for the gene being downloaded.
        """
        # Get the gene's experiment id
        exp_ids = self.get_gene_experiments(gene)

        if exp_ids is None:
            return

        # download experiment data
        for eid in exp_ids:
            print(f"Downloading data for {gene} - experiment: {eid}")
            url = self.download_url.replace("EXP_ID", str(eid))
            download_and_cache(
                url, os.path.join(self.gene_expression_cache, f"{gene}-{eid}")
            )

    def get_gene_data(self, gene, exp_id, use_cache=True, metric="energy"):
        """
        Given a list of gene ids
        """
        logger.debug(f"Getting gene data for gene: {gene} experiment {exp_id}")
        self.gene_name = self.gene_name or gene

        # Check if gene-experiment cached
        if use_cache:
            cache = check_gene_cached(self.gene_expression_cache, gene, exp_id)
        else:
            cache = False

        if not cache:  # then download it
            self.download_gene_data(gene)
            cache = check_gene_cached(self.gene_expression_cache, gene, exp_id)
            if not cache:
                raise ValueError(  # pragma: no cover
                    "Something went wrong and data were not cached"
                )

        # Load from cache
        data = load_cached_gene(cache, metric, self.grid_size)

        if sys.platform == "darwin":
            data = data.T

        return data

    def griddata_to_volume(
        self,
        griddata,
        min_quantile=None,
        min_value=None,
        cmap="bwr",
    ):
        """
        Takes a 3d numpy array with volumetric gene expression
        and returns a vedo.Volume.isosurface actor.
        The isosurface needs a lower bound threshold, this can be
        either a user defined hard value (min_value) or the value
        corresponding to some percentile of the gene expression data.

        :param griddata: np.ndarray, 3d array with gene expression data
        :param min_quantile: float, percentile for threshold
        :param min_value: float, value for threshold
        """
        return Volume(
            griddata,
            min_quantile=min_quantile,
            voxel_size=self.voxel_size,
            min_value=min_value,
            cmap=cmap,
            name=self.gene_name,
            br_class="Gene Data",
        )



import numpy as np
from vedo import Spheres, Sphere
from vedo import Points as vPoints
from pathlib import Path
from pyinspect.utils import _class_name
from loguru import logger


from brainrender.actor import Actor


class Point(Actor):
    def __init__(
        self, pos, radius=100, color="blackboard", alpha=1, res=25, name=None
    ):
        """
        Creates an actor representing a single point
        :param pos: list or np.ndarray with coordinates
        :param radius: float
        :param color: str,
        :param alpha: float
        :param res: int, resolution of mesh
        :param name: str, actor name
        """
        logger.debug(f"Creating a point actor at: {pos}")
        mesh = Sphere(pos=pos, r=radius, c=color, alpha=alpha, res=res)
        name = name or "Point"
        Actor.__init__(self, mesh, name=name, br_class="Point")


class PointsBase:
    def __init__(
        self,
    ):
        """
        Base class with functinality to load from file.
        """
        return

    def _from_numpy(self, data):
        """
        Creates the mesh
        """
        N = len(data)
        if not isinstance(self.colors, str):
            if not N == len(self.colors):  # pragma: no cover
                raise ValueError(  # pragma: no cover
                    "When passing a list of colors, the number of colors shou  # pragma: no coverld match the number of cells"  # pragma: no cover
                )  # pragma: no cover

        self.name = self.name or "Points"
        mesh = Spheres(
            data, r=self.radius, c=self.colors, alpha=self.alpha, res=self.res
        )
        return mesh

    def _from_file(self, data, colors="salmon", alpha=1):
        """
        Loads points coordinates from a numpy file
        before creating the mesh.
        """
        path = Path(data)
        if not path.exists():
            raise FileExistsError(f"File {data} does not exist")

        if path.suffix == ".npy":
            self.name = self.name or path.name
            return self._from_numpy(
                np.load(path),
            )
        else:  # pragma: no cover
            raise NotImplementedError(  # pragma: no cover
                f"Add points from file only works with numpy file for now, not {path.suffix}."  # pragma: no cover
                + "If youd like more formats supported open an issue on Github!"  # pragma: no cover
            )  # pragma: no cover


class Points(PointsBase, Actor):
    def __init__(
        self, data, name=None, colors="salmon", alpha=1, radius=20, res=8
    ):
        """
        Creates an actor representing multiple points (more efficient than
        creating many Point instances).

        :param data: np.ndarray, Nx3 array or path to .npy file with coords data
        :param radius: float
        :param color: str, or list of str with color names or hex codes
        :param alpha: float
        :param name: str, actor name
        :param res: int. Resolution of sphere actors
        """
        PointsBase.__init__(self)
        logger.debug("Creating a Points actor")

        self.radius = radius
        self.colors = colors
        self.alpha = alpha
        self.name = name
        self.res = res

        if isinstance(data, np.ndarray):
            mesh = self._from_numpy(data)
        elif isinstance(data, (str, Path)):
            mesh = self._from_file(data)
        else:  # pragma: no cover
            raise TypeError(  # pragma: no cover
                f"Input data should be either a numpy array or a file path, not: {_class_name(data)}"  # pragma: no cover
            )  # pragma: no cover

        Actor.__init__(self, mesh, name=self.name, br_class="Points")


class PointsDensity(Actor):
    def __init__(
        self, data, name=None, dims=(40, 40, 40), radius=None, **kwargs
    ):
        """
        Creates a Volume actor showing the 3d density of a set
        of points.

        :param data: np.ndarray, Nx3 array with cell coordinates


        from vedo:
            Generate a density field from a point cloud. Input can also be a set of 3D coordinates.
            Output is a ``Volume``.
            The local neighborhood is specified as the `radius` around each sample position (each voxel).
            The density is expressed as the number of counts in the radius search.

            :param int,list dims: numer of voxels in x, y and z of the output Volume.

        """
        logger.debug("Creating a PointsDensity actor")

        # flip coordinates on XY axis to match brainrender coordinates system
        data[:, 2] = -data[:, 2]

        # create volume and then actor
        volume = (
            vPoints(data)
            .density(dims=dims, radius=radius, **kwargs)
            .c("Dark2")
            .alpha([0, 0.9])
            .mode(1)
        )  # returns a vedo Volume
        volume.mirror("z")
        Actor.__init__(self, volume, name=name, br_class="density")



import pyinspect as pi
from io import StringIO
from rich.console import Console
import numpy as np
from myterial import orange, salmon, amber
from vedo import Text3D, Sphere


from brainrender._utils import listify


# transform matrix to fix labels orientation
label_mtx = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]


def make_actor_label(
    atlas,
    actors,
    labels,
    size=300,
    color=None,
    radius=100,
    xoffset=0,
    yoffset=-500,
    zoffset=0,
):
    """
    Adds a 2D text ancored to a point on the actor's mesh
    to label what the actor is

    :param kwargs: key word arguments can be passed to determine
            text appearance and location:
                - size: int, text size. Default 300
                - color: str, text color. A list of colors can be passed
                        if None the actor's color is used. Default None.
                - xoffset, yoffset, zoffset: integers that shift the label position
                - radius: radius of sphere used to denote label anchor. Set to 0 or None to hide.
    """
    offset = [-yoffset, -zoffset, xoffset]
    default_offset = np.array([0, -200, 100])

    new_actors = []
    for n, (actor, label) in enumerate(zip(listify(actors), listify(labels))):

        # Get label color
        if color is None:
            color = [0.2, 0.2, 0.2]

        # Get mesh's highest point
        points = actor.mesh.points().copy()
        point = points[np.argmin(points[:, 1]), :]
        point += np.array(offset) + default_offset
        point[2] = -point[2]

        try:
            if atlas.hemisphere_from_coords(point, as_string=True) == "left":
                point = atlas.mirror_point_across_hemispheres(point)
        except IndexError:
            pass

        # Create label
        txt = Text3D(
            label, point * np.array([1, 1, -1]), s=size, c=color, depth=0.1
        )
        new_actors.append(
            txt.rotateX(180, locally=True).rotateY(180, locally=True)
        )

        # Mark a point on Mesh that corresponds to the label location
        if radius is not None:
            pt = actor.closestPoint(point)
            pt[2] = -pt[2]
            sphere = Sphere(pt, r=radius, c=color, res=8)
            sphere.ancor = pt
            new_actors.append(sphere)
            sphere.computeNormals()

    return new_actors


class Actor(object):
    _needs_label = False  # needs to make a label
    _needs_silhouette = False  # needs to make a silhouette
    _is_transformed = False  # has been transformed to correct axes orientation
    _is_added = False  # has the actor been added to the scene already

    labels = []
    silhouette = None

    def __init__(
        self,
        mesh,
        name=None,
        br_class=None,
        is_text=False,
        color=None,
        alpha=None,
    ):
        """
        Actor class representing anythng shown in a brainrender scene.
        Methods in brainrender.actors are used to creates actors specific
        for different data types.

        An actor has a mesh, a name and a brainrender class type.
        It also has methods to create a silhouette or a label.

        :param mesh: instance of vedo.Mesh
        :param name: str, actor name
        :param br_class: str, name of brainrende actors class
        :param is_text: bool, is it a 2d text or annotation?
        :param color: str, name or hex code of color to assign to actor's mesh
        :param alpha: float, transparency to assign to actor's mesh
        """
        self.mesh = mesh
        self.name = name or "Actor"
        self.br_class = br_class or "None"
        self.is_text = is_text

        if color:
            self.mesh.c(color)
        if alpha:
            self.mesh.alpha(alpha)

    def __getattr__(self, attr):
        """
        If an unknown attribute is called, try `self.mesh.attr`
        to get the meshe's attribute
        """
        if "mesh" not in self.__dict__.keys():
            raise AttributeError(
                f"Actor doesn not have attribute {attr}"
            )  # pragma: no cover

        # some attributes should be from .mesh, others from ._mesh
        mesh_attributes = ("centerOfMass",)
        if attr in mesh_attributes:
            if hasattr(self.__dict__["mesh"], attr):
                return getattr(self.__dict__["mesh"], attr)
        else:
            try:
                return getattr(self.__dict__["_mesh"], attr)
            except KeyError:
                # no ._mesh, use .mesh
                if hasattr(self.__dict__["mesh"], attr):
                    return getattr(self.__dict__["mesh"], attr)

        raise AttributeError(
            f"Actor doesn not have attribute {attr}"
        )  # pragma: no cover

    def __repr__(self):  # pragma: no cover
        return f"brainrender.Actor: {self.name}-{self.br_class}"

    def __str__(self):
        buf = StringIO()
        _console = Console(file=buf, force_jupyter=False)
        _console.print(self)

        return buf.getvalue()

    @property
    def center(self):
        """ returns the coordinates of the mesh's center """
        return self.mesh.points().mean(axis=0)

    @classmethod
    def make_actor(cls, mesh, name, br_class):
        """
        Make an actor from a given mesh
        """
        return cls(mesh, name=name, br_class=br_class)

    def make_label(self, atlas):
        """
        Create a new Actor with a sphere and a text
        labelling this actor
        """
        labels = make_actor_label(
            atlas, self, self._label_str, **self._label_kwargs
        )
        self._needs_label = False

        lbls = [
            Actor.make_actor(label, self.name, "label") for label in labels
        ]
        self.labels = lbls
        return lbls

    def make_silhouette(self):
        """
        Create a new silhouette actor outlining this actor
        """
        lw = self._silhouette_kwargs["lw"]
        color = self._silhouette_kwargs["color"]
        sil = self._mesh.silhouette().lw(lw).c(color)

        name = f"{self.name} silhouette"
        sil = Actor.make_actor(sil, name, "silhouette")
        sil._is_transformed = True

        self._needs_silhouette = False
        self.silhouette = sil

        return sil

    def __rich_console__(self, *args):
        """
        Print some useful characteristics to console.
        """
        rep = pi.Report(
            title="[b]brainrender.Actor: ",
            color=salmon,
            accent=orange,
        )

        rep.add(f"[b {orange}]name:[/b {orange}][{amber}] {self.name}")
        rep.add(f"[b {orange}]type:[/b {orange}][{amber}] {self.br_class}")
        rep.line()
        rep.add(
            f"[{orange}]center of mass:[/{orange}][{amber}] {self.mesh.centerOfMass().astype(np.int32)}"
        )
        rep.add(
            f"[{orange}]number of vertices:[/{orange}][{amber}] {len(self.mesh.points())}"
        )
        rep.add(
            f"[{orange}]dimensions:[/{orange}][{amber}] {np.array(self.mesh.bounds()).astype(np.int32)}"
        )
        rep.add(f"[{orange}]color:[/{orange}][{amber}] {self.mesh.color()}")

        yield "\n"
        yield rep



from pathlib import Path


def listdir(fld):
    """
    List the files into a folder with the complete file path instead of the relative file path like os.listdir.

    :param fld: string, folder path

    """
    return [str(f) for f in Path(fld).glob("**/*") if f.is_file()]


def get_subdirs(folderpath):
    """
    Returns the subfolders in a given folder
    """
    return [str(f) for f in Path(folderpath).glob("**/*") if f.is_dir()]


def listify(obj):
    """
    makes sure that the obj is a list
    """
    if isinstance(obj, list):
        return obj
    elif isinstance(obj, tuple):
        return list(obj)
    else:
        return [obj]


def return_list_smart(lst):
    """
    If the list has length > returns the list
    if it has length == 1 it returns the element
    if it has length == 0 it returns None
    """
    if len(lst) > 1:
        return lst
    elif len(lst) == 1:
        return lst[0]
    else:
        return None



from vedo import Video as VtkVideo
from myterial import amber_light
from rich import print
from loguru import logger
import os


class Video(VtkVideo):
    # Redifine vedo.Video close method
    def __init__(self, *args, fmt="mp4", size="1620x1050", **kwargs):
        """
        Video class, takes care of storing screenshots (frames)
        as images in a temporary folder and then merging these into a
        single video file when the video is closed.
        """
        super().__init__(*args, **kwargs)
        self.format = fmt
        self.size = size

    def close(self):
        """Render the video and write to file."""
        print(f"[{amber_light}]Saving video")
        logger.debug(f"[{amber_light}]Saving video")

        fld = os.path.join(self.tmp_dir.name, "%d.png")
        fps = int(self.fps)
        name = f"{self.name}.{self.format}"
        fmt = "-vcodec libx264 -crf 28 -pix_fmt yuv420p"
        if self.size:
            fmt += f" -s {self.size}"

        command = f"ffmpeg -hide_banner -loglevel panic -y -r {fps} -start_number 0 -i {fld} {fmt} {name}"
        out = os.system(command)
        self.tmp_dir.cleanup()
        return out, command



from vedo import Plotter
from collections import namedtuple
import datetime
from loguru import logger
from qtpy.QtWidgets import QFrame
from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

import brainrender
from brainrender import Scene
from brainrender.camera import set_camera

from brainrender.gui.ui import UI
from brainrender.gui.apputils.camera_control import CameraControl
from brainrender.gui.apputils.add_from_file_control import AddFromFile
from brainrender.gui.apputils.regions_control import RegionsControl
from brainrender.gui.apputils.actors_control import ActorsControl

from brainrender.gui.widgets.actors_list import update_actors_list
from brainrender.gui.widgets.screenshot_modal import ScreenshotModal


class App(
    Scene, UI, CameraControl, AddFromFile, RegionsControl, ActorsControl
):
    startup = True  # some things only run once
    actors = {}  # stores actors and status
    camera_orientation = None  # used to manually set camera orientation

    def __init__(self, *args, atlas_name=None, axes=None, **kwargs):
        """
        Initialise the qtpy app and the brainrender scene.

        Arguments:
        ----------

        atlas_name: str/None. Name of the brainglobe atlas to use
        axes: bool. If true axes are shown in the brainrender render
        """
        logger.debug("Creating brainrender GUI")

        # make a vtk widget for the vedo plotter
        frame = QFrame()
        self.vtkWidget = QVTKRenderWindowInteractor(frame)

        # Get vtkWidget plotter and creates a scene embedded in it
        new_plotter = Plotter(qtWidget=self.vtkWidget)
        self.scene = Scene(
            *args, atlas_name=atlas_name, plotter=new_plotter, **kwargs
        )

        # Initialize parent classes
        UI.__init__(self, *args, **kwargs)
        CameraControl.__init__(self)
        AddFromFile.__init__(self)
        RegionsControl.__init__(self)
        ActorsControl.__init__(self)

        # Setup brainrender plotter
        self.axes = axes
        self.atuple = namedtuple("actor", "mesh, is_visible, color, alpha")

        self.setup_scene()
        self._update()
        self.scene.render()
        self.scene._get_inset()

        # Setup widgets functionality
        self.actors_list.itemDoubleClicked.connect(
            self.actor_list_double_clicked
        )
        self.actors_list.clicked.connect(self.actor_list_clicked)

        buttons_funcs = dict(
            add_brain_regions=self.open_regions_dialog,
            add_from_file=self.add_from_file_object,
            add_cells=self.add_from_file_cells,
            show_structures_tree=self.toggle_treeview,
            take_screenshot=self.take_screenshot,
            reset=self.reset_camera,
            top=self.move_camera_top,
            side1=self.move_camera_side1,
            side2=self.move_camera_side2,
            front=self.move_camera_front,
        )

        for btn, fun in buttons_funcs.items():
            self.buttons[btn].clicked.connect(fun)

        self.treeView.clicked.connect(self.add_region_from_tree)

        self.alpha_textbox.textChanged.connect(self.update_actor_properties)
        self.color_textbox.textChanged.connect(self.update_actor_properties)

        self.startup = False

    def take_screenshot(self):
        logger.debug("GUI: taking screenshot")
        self._update()
        self.scene.plotter.render()

        # Get savename
        self.scene.screenshots_folder.mkdir(exist_ok=True)

        savename = str(self.scene.screenshots_folder / "brainrender_gui")
        savename += (
            f'_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}' + ".png"
        )
        print(f"\nSaving screenshot at {savename}\n")
        self.scene.screenshot(name=savename)

        # show success message
        dialog = ScreenshotModal(self, self.palette)
        dialog.exec()

    # ------------------------------ Toggle treeview ----------------------------- #
    def toggle_treeview(self):
        """
        Method for the show structures tree button.
        It toggles the visibility of treeView widget
        and adjusts the button's text accordingly.
        """
        logger.debug("GUI: toggle tree view")
        if not self.treeView.isHidden():
            self.buttons["show_structures_tree"].setText(
                "Show structures tree"
            )
        else:
            self.buttons["show_structures_tree"].setText(
                "Hide structures tree"
            )

        self.treeView.setHidden(not self.treeView.isHidden())

    # ------------------------------- Initial setup ------------------------------ #
    def setup_scene(self):
        """
        Set scene's axes and camera
        """
        # Get axes
        if self.axes:
            self.axes = self._make_axes()
            # self.scene.add_actor(ax)
        else:
            self.axes = None

        # Fix camera
        set_camera(self.scene, self.scene.plotter.camera)

    # ---------------------------------- Update ---------------------------------- #
    def _update_actors(self):
        """
        All actors that are part of the scene are stored
        in a dictionary with key as the actor name and
        value as a 4-tuple with (Mesh, is_visible, color, alpha).
        `is_visible` is a bool that determines if the
        actor should be rendered
        """

        for actor in self.scene.actors:
            if actor is None:
                continue

            try:
                if actor.name not in self.actors.keys():
                    self.actors[actor.name] = self.atuple(
                        actor, True, actor.mesh.color(), actor.mesh.alpha()
                    )

                    if actor.silhouette is not None:
                        self.scene.plotter.remove(actor.silhouette.mesh)
                        actor.make_silhouette()

                        self.scene.plotter.add(actor.silhouette.mesh)
                        self.actors[actor.silhouette.name] = self.atuple(
                            actor.silhouette,
                            True,
                            actor.silhouette.mesh.color(),
                            actor.silhouette.mesh.alpha(),
                        )
            except AttributeError:
                # the Assembly object representing the axes should be ignore
                pass

    def _update(self):
        """
        Updates the scene's Plotter to add/remove
        meshes
        """

        # update meshes
        self.scene._apply_style()

        # Get actors to render
        self._update_actors()
        to_render = [act for act in self.actors.values() if act.is_visible]

        # Set actors look
        meshes = [act.mesh.c(act.color).alpha(act.alpha) for act in to_render]

        # Add axes
        if self.axes is not None:
            meshes.append(self.axes)

        # update actors rendered
        self.scene.plotter.show(
            *meshes,
            interactorStyle=0,
            bg=brainrender.settings.BACKGROUND_COLOR,
            resetcam=self.startup,
            zoom=None,
        )

        # Fake a button press to force canvas update
        self.scene.plotter.interactor.MiddleButtonPressEvent()
        self.scene.plotter.interactor.MiddleButtonReleaseEvent()

        # Update list widget
        update_actors_list(self.actors_list, self.actors)
        return meshes

    # ----------------------------------- Close ---------------------------------- #
    def onClose(self):
        """
        Disable the interactor before closing to prevent it from trying to act on a already deleted items
        """
        self.vtkWidget.close()



from pathlib import Path
from vedo import load
import requests


def connected_to_internet(url="http://www.google.com/", timeout=5):
    """
    Check that there is an internet connection

    :param url: url to use for testing (Default value = 'http://www.google.com/')
    :param timeout:  timeout to wait for [in seconds] (Default value = 5)
    """

    try:
        _ = requests.get(url, timeout=timeout)
        return True
    except requests.ConnectionError:  # pragma: no cover
        print("No internet connection available.")  # pragma: no cover
    return False


def fail_on_no_connection(func):
    """
    Decorator that throws an error if no internet connection is available
    """
    if not connected_to_internet():  # pragma: no cover
        raise ConnectionError(
            "No internet connection found."
        )  # pragma: no cover

    def inner(*args, **kwargs):
        return func(*args, **kwargs)

    return inner


def request(url):
    """
    Sends a request to a url

    :param url:

    """
    if not connected_to_internet():  # pragma: no cover
        raise ConnectionError(
            "No internet connection found."
        )  # pragma: no cover

    response = requests.get(url)
    if response.ok:
        return response
    else:  # pragma: no cover
        exception_string = "URL request failed: {}".format(
            response.reason
        )  # pragma: no cover
    raise ValueError(exception_string)


def check_file_exists(func):  # pragma: no cover
    """
    Decorator that throws an error if a function;s first argument
    is not a path to an existing file.
    """

    def inner(*args, **kwargs):
        if not Path(args[0]).exists():
            raise FileNotFoundError(
                f"File {args[0]} not found"
            )  # pragma: no cover
        return func(*args, **kwargs)

    return inner


@check_file_exists
def load_mesh_from_file(filepath, color=None, alpha=None):
    """
    Load a a mesh or volume from files like .obj, .stl, ...

    :param filepath: path to file
    :param **kwargs:

    """
    actor = load(str(filepath))
    actor.c(color).alpha(alpha)
    return actor



"""
Scene
    - Create a scene, add root and inset if necessary
    - add actor method
    - special methods

"""
import sys
from pathlib import Path
from vedo import Mesh, Text2D, Assembly
import pyinspect as pi
from rich import print
from loguru import logger
from myterial import amber, orange, orange_darker, salmon

from brainrender import settings
from brainrender.atlas import Atlas
from brainrender.render import Render
from brainrender.actor import Actor
from brainrender.actors import Volume
from brainrender._utils import return_list_smart, listify
from brainrender._io import load_mesh_from_file
from brainrender._jupyter import not_on_jupyter, JupyterMixIn


class Scene(JupyterMixIn, Render):
    def __init__(
        self,
        root=True,
        atlas_name=None,
        inset=True,
        title=None,
        screenshots_folder=None,
        plotter=None,
        title_color: str = "k",
    ):
        """
        Main scene in brainrender.
        It coordinates what should be render and how should it look like.

        :param root: bool. If true the brain root mesh is added
        :param atlas_name: str, name of the brainglobe atlas to be used
        :param inset: bool. If true an inset is shown with the brain's outline
        :param title: str. If true a title is added to the top of the window
        :param screenshots_folder: str, Path. Where the screenshots will be saved
        """
        logger.debug(
            f"Creating scene with parameters: root: {root}, atlas_name: '{atlas_name}'', inset: {inset}, screenshots_folder: {screenshots_folder}"
        )
        JupyterMixIn.__init__(self)

        self.actors = []  # stores all actors in the scene
        self.labels = []  # stores all `labels` actors in scene

        self.atlas = Atlas(atlas_name=atlas_name)

        self.screenshots_folder = (
            Path(screenshots_folder)
            if screenshots_folder is not None
            else Path().cwd()
        )
        self.screenshots_folder.mkdir(exist_ok=True)

        # Initialise render class
        Render.__init__(self, plotter)

        # Get root mesh
        self.root = self.add_brain_region(
            "root",
            alpha=settings.ROOT_ALPHA,
            color=settings.ROOT_COLOR,
            silhouette=True
            if root and settings.SHADER_STYLE == "cartoon"
            else False,
        )
        self.atlas.root = self.root  # give atlas access to root
        self._root_mesh = self.root.mesh.clone()
        if not root:
            self.remove(self.root)

        # keep track if we need to make an inset
        self.inset = inset

        # add title
        if title:
            self.add(
                Text2D(title, pos="top-center", s=2.5, c=title_color, alpha=1),
                names="title",
                classes="title",
            )

    def __str__(self):
        return f"A `brainrender.scene.Scene` with {len(self.actors)} actors."

    def __repr__(self):  # pragma: no cover
        return f"A `brainrender.scene.Scene` with {len(self.actors)} actors."

    def __repr_html__(self):  # pragma: no cover
        return f"A `brainrender.scene.Scene` with {len(self.actors)} actors."

    def __del__(self):
        self.close()

    @not_on_jupyter
    def _get_inset(self):
        """
        Creates a small inset showing the brain's orientation
        """
        if settings.OFFSCREEN:
            return None

        inset = self._root_mesh.clone()
        inset.alpha(1)  # scale(0.5)
        self.plotter.addInset(inset, pos=(0.95, 0.1), draggable=False)

        if settings.SHADER_STYLE == "cartoon":
            inset.lighting("off")

    def add(self, *items, names=None, classes=None, transform=True, **kwargs):
        """
        General method to add Actors to the scene.

        :param items: vedo.Mesh, Actor, (str, Path).
                If str/path it should be a path to a .obj or .stl file.
                Whatever the input it's turned into an instance of Actor
                before adding it to the scne

        :param names: names to be assigned to the Actors
        :param classs: br_classes to be assigned to the Actors
        :param **kwargs: parameters to be passed to the individual
            loading functions (e.g. to load from file and specify the color)
        """
        names = names or [None for a in items]
        classes = classes or [None for a in items]

        # turn items into Actors
        actors = []
        for item, name, _class in zip(items, listify(names), listify(classes)):
            if item is None:
                continue

            if isinstance(item, (Mesh, Assembly)):
                actors.append(Actor(item, name=name, br_class=_class))

            elif isinstance(item, Text2D):
                # Mark text actors differently because they don't behave like
                # other 3d actors
                actors.append(
                    Actor(
                        item,
                        name=name,
                        br_class=_class,
                        is_text=True,
                        **kwargs,
                    )
                )
            elif pi.utils._class_name(item) == "Volume" and not isinstance(
                item, Volume
            ):
                actors.append(
                    Volume(item, name=name, br_class=_class, **kwargs)
                )
            elif isinstance(item, Actor):
                actors.append(item)

            elif isinstance(item, (str, Path)):
                mesh = load_mesh_from_file(item, **kwargs)
                name = name or Path(item).name
                _class = _class or "from file"
                actors.append(Actor(mesh, name=name, br_class=_class))

            else:
                raise ValueError(
                    f"Unrecognized argument: {item} [{pi.utils._class_name(item)}]"
                )

        # transform actors
        if transform:
            for actor in actors:
                self._prepare_actor(actor)

        # add actors to plotter
        for actor in actors:
            try:
                self.plotter.add(actor._mesh)
            except AttributeError:  # e.g. for titles
                self.plotter.add(actor.mesh)

        # Add to the lists actors
        self.actors.extend(actors)
        return return_list_smart(actors)

    def remove(self, *actors):
        """
        Removes actors from the scene.
        """
        logger.debug(f"Removing {len(actors)} actors from scene")
        for act in actors:
            try:
                self.actors.pop(self.actors.index(act))
            except Exception:
                print(
                    f"Could not remove ({act}, {pi.utils._class_name(act)}) from actors"
                )
            else:
                # remove from plotter
                try:
                    self.plotter.remove(act._mesh)
                except AttributeError:
                    pass

                if act.silhouette is not None:
                    self.plotter.remove(act.silhouette.mesh)

                for label in act.labels:
                    self.plotter.remove(label.mesh)

    def get_actors(self, name=None, br_class=None):
        """
        Return's the scene's actors that match some search criteria.

        :param name: strm int or list of str/int, actors' names
        :param br_class: str or list of str, actors br classes
        """
        matches = self.actors
        if name is not None:
            name = listify(name)
            matches = [m for m in matches if m.name in name]
        if br_class is not None:
            br_class = listify(br_class)
            matches = [m for m in matches if m.br_class in br_class]
        return matches

    def add_brain_region(
        self,
        *regions,
        alpha=1,
        color=None,
        silhouette=None,
        hemisphere="both",
        force=False,
    ):
        """
        Dedicated method to add brain regions to render

        :param regions: str. String of regions names
        :param alpha: float
        :param color: str. If None the atlas default color is used
        :param silhouette: bool. If true regions Actors will have
            a silhouette
        :param hemisphere: str.
            - if "both" the complete mesh is returned
            - if "left"/"right" only the corresponding half
                of the mesh is returned
        :param force: force adding of region even if already rendred
        """
        if silhouette is None:
            silhouette = (
                silhouette or True
                if settings.SHADER_STYLE == "cartoon"
                else False
            )

        # avoid adding regions already rendered
        if not force:
            already_in = [
                r.name for r in self.get_actors(br_class="brain region")
            ]
            regions = [r for r in regions if r not in already_in]

        if not regions:  # they were all already rendered
            logger.debug(
                "Not adding any region because they are all already in the scene"
            )
            return None

        logger.debug(
            f"SCENE: Adding {len(regions)} brain regions to scene: {regions}"
        )

        # get regions actors from atlas
        regions = self.atlas.get_region(*regions, alpha=alpha, color=color)
        regions = listify(regions) or []

        # add actors
        actors = self.add(*regions)

        # slice to keep only one hemisphere
        if hemisphere == "right":
            plane = self.atlas.get_plane(
                pos=self.root._mesh.centerOfMass(), norm=(0, 0, 1)
            )
        elif hemisphere == "left":
            plane = self.atlas.get_plane(
                pos=self.root._mesh.centerOfMass(), norm=(0, 0, -1)
            )

        if hemisphere in ("left", "right"):
            if not isinstance(actors, list):
                actors._mesh.cutWithPlane(
                    origin=plane.center,
                    normal=plane.normal,
                )
                actors.cap()
            else:
                for actor in actors:
                    actor._mesh.cutWithPlane(
                        origin=plane.center,
                        normal=plane.normal,
                    )
                    actor.cap()

        # make silhouettes
        if silhouette and regions and alpha:
            self.add_silhouette(*regions, lw=2)

        return actors

    @not_on_jupyter
    def add_silhouette(self, *actors, lw=1, color="k"):
        """
        Dedicated method to add silhouette to actors

        :param actors: Actors
        :param lw: float. Line weight
        :param color: str, silhouette color
        """
        for actor in actors:
            if actor is None:
                continue
            actor._needs_silhouette = True
            actor._silhouette_kwargs = dict(
                lw=lw or settings.LW,
                color=color,
            )

    @not_on_jupyter
    def add_label(self, actor, label, **kwargs):
        """
        Dedicated method to add lables to actors

        :param actor: Actors
        :param llabelw: str. Text of label
        :param **kwargs: see brainrender._actor.make_actor_label for kwargs
        """
        actor._needs_label = True
        actor._label_str = label
        actor._label_kwargs = kwargs

    def slice(
        self,
        plane,
        actors=None,
        close_actors=False,
    ):
        """
        Slices actors with a plane.

        :param plane: str, Plane. If a string it needs to be
            a supported plane from brainglobe's atlas api (e.g. 'frontal')
            otherwise it should be a vedo.Plane mesh
        :param actors: list of actors to be sliced. If None all actors
            will be sliced
        :param close_actors: If true the openings in the actors meshes
            caused by teh cut will be closed.
        """
        if isinstance(plane, str):
            plane = self.atlas.get_plane(plane=plane)

        if not actors or actors is None:
            actors = self.clean_actors.copy()

        for actor in listify(actors):
            actor._mesh = actor._mesh.cutWithPlane(
                origin=plane.center,
                normal=plane.normal,
            )
            if close_actors:
                actor.cap()

            if actor.silhouette is not None:
                self.plotter.remove(actor.silhouette.mesh)
                self.plotter.add(actor.make_silhouette().mesh)

    @property
    def content(self):
        """
        Prints an overview of the Actors in the scene.
        """

        actors = pi.Report(
            "Scene actors", accent=salmon, dim=orange, color=orange
        )

        for act in self.actors:
            actors.add(
                f"[bold][{amber}]- {act.name}[/bold][{orange_darker}] (type: [{orange}]{act.br_class}[/{orange}])"
            )

        if "win32" != sys.platform:
            actors.print()
        else:
            print(pi.utils.stringify(actors, maxlen=-1))

    @property
    def renderables(self):
        """
        Returns the meshes for all actors.
        """
        if not self.backend:
            return [a.mesh for a in self.actors + self.labels]
        else:
            return [a.mesh for a in self.actors if not a.is_text]

    @property
    def clean_actors(self):
        """
        returns only ators that are not Text objects and similar
        """
        return [a for a in self.actors if not a.is_text]

    @property
    def clean_renderables(self):
        """
        Returns meshses only for 'clean actors' (i.e. not text)
        """
        return [a.mesh for a in self.actors if not a.is_text]



from vedo import Plotter
from vedo import settings as vsettings
import numpy as np
from datetime import datetime
from rich import print
from pathlib import Path
from myterial import orange, amber, deep_purple_light, teal
from rich.syntax import Syntax
from loguru import logger

from brainrender import settings
from brainrender.camera import (
    get_camera,
    check_camera_param,
    set_camera,
    get_camera_params,
)
from brainrender.actors.points import PointsDensity


# mtx used to transform meshes to sort axes orientation
mtx = [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]


class Render:
    is_rendered = False
    plotter = None

    axes_names = ("AP", "DV", "LR")
    axes_lookup = {"x": "AP", "y": "DV", "z": "LR"}
    axes_indices = {"AP": 0, "DV": 1, "LR": 2}

    def __init__(self, plotter=None):
        """
        Backend for Scene, handles all rendering and exporting
        related tasks.
        """
        if plotter is None:
            self._get_plotter()
        else:
            self.plotter = plotter
            self.plotter.keyPressFunction = self.keypress

    def _get_plotter(self):
        """
        Make a vedo plotter with
        fancy axes and all
        """
        self.plotter = Plotter(
            axes=self._make_axes() if settings.SHOW_AXES else None,
            pos=(0, 0),
            title="brainrender",
            bg=settings.BACKGROUND_COLOR,
            offscreen=settings.OFFSCREEN,
            size="full" if settings.WHOLE_SCREEN else "auto",
        )

        self.plotter.keyPressFunction = self.keypress

    def _make_axes(self):
        """
        Returns a dictionary with axes
        parameters for the vedo plotter
        """
        ax_idx = self.atlas.space.axes_order.index("frontal")

        # make acustom axes dict
        atlas_shape = np.array(self.atlas.metadata["shape"]) * np.array(
            self.atlas.metadata["resolution"]
        )
        z_range = np.array([-atlas_shape[2], 0])
        z_ticks = [
            (-v, str(np.abs(v).astype(np.int32)))
            for v in np.linspace(
                0,
                atlas_shape[ax_idx],
                10,
            )
        ]

        if self.atlas.atlas_name == "allen_human_500um":
            z_range = None
            z_ticks = None
            logger.debug(
                "RENDER: manually forcing axes size for human atlas, atlas needs fixing"
            )

        # make custom axes dict
        axes = dict(
            axesLineWidth=3,
            tipSize=0,
            xtitle="AP (μm)",
            ytitle="DV (μm)",
            ztitle="LR (μm)",
            textScale=0.8,
            xTitleRotation=180,
            zrange=z_range,
            zValuesAndLabels=z_ticks,
            xyGrid=False,
            yzGrid=False,
            zxGrid=False,
            xUseBounds=True,
            yUseBounds=True,
            zUseBounds=True,
            xLabelRotation=180,
            yLabelRotation=180,
            zLabelRotation=90,
        )

        return axes

    def _prepare_actor(self, actor):
        """
        When an actor is first rendered, a transform matrix
        is applied to its points to correct axes orientation
        mismatches: https://github.com/brainglobe/bg-atlasapi/issues/73

        Once an actor is 'corrected' it spawns labels and silhouettes as needed
        """
        # don't apply transforms to points density actors
        if isinstance(actor, PointsDensity):
            logger.debug(
                f'Not transforming actor "{actor.name} (type: {actor.br_class})"'
            )
            actor._is_transformed = True

        # Flip every actor's orientation
        if not actor._is_transformed:
            try:
                actor._mesh = actor.mesh.clone()
                actor._mesh.applyTransform(mtx)
            except AttributeError:  # some types of actors dont trasform
                logger.debug(
                    f'Failed to transform actor: "{actor.name} (type: {actor.br_class})"'
                )
                actor._is_transformed = True
            else:
                try:
                    actor.mesh.reverse()
                except AttributeError:  # Volumes don't have reverse
                    logger.debug(
                        f'Failed to reverse actor: "{actor.name} (type: {actor.br_class})"'
                    )
                    pass
                actor._is_transformed = True

        # Add silhouette and labels
        if actor._needs_silhouette and not self.backend:
            self.plotter.add(actor.make_silhouette().mesh)

        if actor._needs_label and not self.backend:
            self.labels.extend(actor.make_label(self.atlas))

    def _apply_style(self):
        """
        Sets the rendering style for each mesh
        """
        for actor in self.clean_actors:
            if settings.SHADER_STYLE != "cartoon":
                style = settings.SHADER_STYLE
            else:
                if self.backend:  # notebook backend
                    print(
                        'Shader style "cartoon" cannot be used in a notebook'
                    )
                style = "off"

            try:
                actor.mesh.reverse()  # flip normals
                actor.mesh.lighting(style=style)

                actor._mesh.reverse()
                actor._mesh.lighting(style=style)
            except AttributeError:
                pass

    def render(
        self,
        interactive=None,
        camera=None,
        zoom=None,
        update_camera=True,
        **kwargs,
    ):
        """
        Renders the scene.

        :param interactive: bool. If note settings.INTERACTIVE is used.
            If true the program's execution is stopped and users
            can interact with scene.
        :param camera: str, dict. If none the default camera is used.
            Pass a valid camera input to specify the camera position when
            the scene is rendered.
        :param zoom: float, if None atlas default is used
        :param update_camera: bool, if False the camera is not changed
        :param kwargs: additional arguments to pass to self.plotter.show
        """
        logger.debug(
            f"Rendering scene. Interactive: {interactive}, camera: {camera}, zoom: {zoom}"
        )
        # get zoom
        zoom = zoom or self.atlas.zoom

        # get vedo plotter
        if self.plotter is None:
            self._get_plotter()

        # Get camera
        camera = camera or settings.DEFAULT_CAMERA
        if isinstance(camera, str):
            camera = get_camera(camera)
        else:
            camera = check_camera_param(camera)

        if "focalPoint" not in camera.keys() or camera["focalPoint"] is None:
            camera["focalPoint"] = self.root._mesh.centerOfMass()

        if not self.backend and camera is not None:
            camera = set_camera(self, camera)

        # Apply axes correction
        for actor in self.clean_actors:
            if not actor._is_transformed:
                self._prepare_actor(actor)
                self.plotter.add(actor.mesh)

            if actor._needs_silhouette or actor._needs_label:
                self._prepare_actor(actor)

        # add labels to the scene
        for label in self.labels:
            if label._is_added:
                continue
            else:
                label._mesh = label.mesh.clone()
                self._prepare_actor(label)
                self.plotter.add(label._mesh.reverse())
                label._is_added = True

        # Apply style
        self._apply_style()

        if self.inset and not self.is_rendered:
            self._get_inset()

        # render
        self.is_rendered = True
        if not self.backend:  # not running in a python script
            if interactive is None:
                interactive = settings.INTERACTIVE

            self.plotter.show(
                interactive=interactive,
                zoom=zoom,
                bg=settings.BACKGROUND_COLOR,
                camera=camera.copy() if update_camera else None,
                interactorStyle=0,
                rate=40,
            )
        elif self.backend == "k3d":  # pragma: no cover
            # Remove silhouettes
            self.remove(*self.get_actors(br_class="silhouette"))
            print(
                f"[{teal}]Your scene is ready for rendering, use:\n",
                Syntax("from vedo import show", lexer_name="python"),
                Syntax("vedo.show(*scene.renderables)", lexer_name="python"),
                sep="\n",
            )
        else:  # pragma: no cover
            print(
                f"[{teal}]Your scene is ready for rendering, use:\n",
                Syntax("from itkwidgets import view", lexer_name="python"),
                Syntax(
                    "view(scene.plotter.show(*scene.renderables))",
                    lexer_name="python",
                ),
                sep="\n",
            )

    def close(self):
        self.plotter.close()

    def export(self, savepath):
        """
        Exports the scene to a .html
        file for online renderings.

        :param savepath: str, Path to a .html file to save the export
        """
        logger.debug(f"Exporting scene to {savepath}")
        _backend = self.backend

        if not self.is_rendered:
            self.render(interactive=False)

        path = Path(savepath)
        if path.suffix != ".html":
            raise ValueError("Savepath should point to a .html file")

        # prepare settings
        vsettings.notebookBackend = "k3d"

        # Create new plotter and save to file
        plt = Plotter()
        plt.add(self.clean_renderables, render=False)
        plt = plt.show(interactive=False)
        plt.camera[-2] = -1

        with open(path, "w") as fp:
            fp.write(plt.get_snapshot())

        print(
            f"The brainrender scene has been exported for web. The results are saved at {path}"
        )

        # Reset settings
        vsettings.notebookBackend = None
        self.backend = _backend

        return str(path)

    def screenshot(self, name=None, scale=None):
        """
        Takes a screenshot of the current view
        and save it to file.
        Screenshots are saved in `screenshots_folder`
        (see Scene)

        :param name: str, name of png file
        :param scale: float, >1 for higher resolution
        """

        if not self.is_rendered:
            self.render(interactive=False)

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        name = name or f"brainrender_screenshot_{timestamp}"
        if ".png" not in name:
            name += ".png"

        scale = scale or settings.SCREENSHOT_SCALE

        print(f"\nSaving new screenshot at {name}\n")

        savepath = str(self.screenshots_folder / name)
        logger.debug(f"Saving scene at {savepath}")
        self.plotter.screenshot(filename=savepath, scale=scale)
        return savepath

    def _print_camera(self):
        pms = get_camera_params(scene=self)

        focal = pms.pop("focalPoint", None)
        dst = pms.pop("distance", None)

        names = [
            f"[green bold]     '{k}'[/green bold]: [{amber}]{v},"
            for k, v in pms.items()
        ]
        print(
            f"[{deep_purple_light}]Camera parameters:",
            f"[{orange}]    {{",
            *names,
            f"[{orange}]   }}",
            f"[{deep_purple_light}]Additional, (optional) parameters:",
            f"[green bold]     'focalPoint'[/green bold]: [{amber}]{focal},",
            f"[green bold]     'distance'[/green bold]: [{amber}]{dst},",
            sep="\n",
        )

    def keypress(self, key):  # pragma: no cover
        """
        Hanles key presses for interactive view
        -s: take's a screenshot
        -q: closes the window
        -c: prints the current camera parameters
        """
        if key == "s":
            self.screenshot()

        elif key == "q" or key == "Esc":
            self.close()

        elif key == "c":
            self._print_camera()



from qtpy.QtGui import QFont, QColor, QStandardItem


class StandardItem(QStandardItem):
    def __init__(self, txt="", tag=None, depth=0, color=None):
        """
        Items in the tree list with some
        extended functionality to specify/update
        their look.
        """
        super().__init__()
        self.depth = depth  # depth in the hierarchy structure
        self.tag = tag

        self._checked = False

        # Set font color/size
        self.toggle_active()

        # Set text
        self.setEditable(False)
        rgb = color.replace(")", "").replace(" ", "").split("(")[-1].split(",")
        self.setForeground(QColor(*[int(r) for r in rgb]))
        self.setText(txt)

        # Set checkbox
        # self.setFlags(
        #     self.flags() | Qt.ItemIsTristate | Qt.ItemIsUserCheckable
        # )
        # self.setCheckState(Qt.Unchecked)

    def toggle_active(self):
        """
        When a mesh corresponding to the item's region
        get's rendered, change the font to bold
        to highlight the fact.
        """
        fnt = QFont("Roboto", 14)
        if self._checked:
            fnt.setBold(True)
        else:
            fnt.setBold(False)
        self.setFont(fnt)



import numpy as np
from loguru import logger

from vedo import merge
from vedo.shapes import Line, Sphere, Text3D
from vedo.utils import precision, mag

from brainrender.actor import Actor


def ruler(p1, p2, unit_scale=1, units=None, s=50):
    """
    Creates a ruler showing the distance between two points.
    The ruler is composed of a line between the points and
    a text indicating the distance.

    :param p1: list, np.ndarray with coordinates of first point
    :param p2: list, np.ndarray with coordinates of second point
    :param unit_scale: float. To scale the units (e.g. show mm instead of µm)
    :param units: str, name of unit (e.g. 'mm')
    :param s: float size of text

    """
    logger.debug(f"Creating a ruler actor between {p1} and {p2}")
    actors = []

    # Make two line segments
    midpoint = np.array([(x + y) / 2 for x, y in zip(p1, p2)])
    gap1 = ((midpoint - p1) * 0.8) + p1
    gap2 = ((midpoint - p2) * 0.8) + p2

    actors.append(Line(p1, gap1, lw=200))
    actors.append(Line(gap2, p2, lw=200))

    # Add label
    if units is None:  # pragma: no cover
        units = ""  # pragma: no cover
    dist = mag(p2 - p1) * unit_scale
    label = precision(dist, 3) + " " + units
    lbl = Text3D(label, pos=midpoint, s=s + 100, justify="center")
    lbl.SetOrientation([0, 0, 180])
    actors.append(lbl)

    # Add spheres add end
    actors.append(Sphere(p1, r=s, c=[0.3, 0.3, 0.3]))
    actors.append(Sphere(p2, r=s, c=[0.3, 0.3, 0.3]))

    act = Actor(merge(*actors), name="Ruler", br_class="Ruler")
    act.c((0.3, 0.3, 0.3)).alpha(1).lw(2)
    return act


def ruler_from_surface(
    p1, root, unit_scale=1, axis=1, units=None, s=50
) -> Actor:
    """
    Creates a ruler between a point and the brain's surface
    :param p1: list, np.ndarray with coordinates of  point
    :param root: mesh or actor with brain's root
    :param axis: int, index of axis along which distance is computed
    :param unit_scale: float. To scale the units (e.g. show mm instead of µm)
    :param units: str, name of unit (e.g. 'mm')
    :param s: float size of text
    """
    logger.debug(f"Creating a ruler actor between {p1} and brain surface")
    # Get point on brain surface
    p2 = p1.copy()
    p2[axis] = 0  # zero the choosen coordinate

    pts = root.mesh.intersectWithLine(p1, p2)
    surface_point = pts[0]

    return ruler(p1, surface_point, unit_scale=unit_scale, units=units, s=s)



from qtpy.QtGui import QIcon
from pkg_resources import resource_filename
from loguru import logger


def get_in_alist(qlist):
    """
    Gets items in the actor list
    """
    items = []
    for index in range(qlist.count()):
        items.append(qlist.item(index).text())
    return items


def update_actors_list(qlist, actorsdict):
    """
    Adds missing entries in the actors list
    """
    listed = get_in_alist(qlist)

    # Add items to list
    for actor in actorsdict.keys():
        if actor not in listed:
            qlist.insertItem(qlist.count() + 1, actor)

            item = qlist.item(qlist.count() - 1)
            item.setIcon(
                QIcon(resource_filename("brainrender.gui", "icons/eye.svg"))
            )


def remove_from_list(qlist, aname):
    """
    Removes an entry from the actors list
    """
    logger.debug(f"GUI: removing {aname} from actors list")
    if aname not in get_in_alist(qlist):
        raise ValueError(
            f"Attempting to remove {aname} from list, but {aname} not in list."
        )
    else:
        idx = [n for n, a in enumerate(get_in_alist(qlist)) if a == aname][0]
        qlist.takeItem(idx)



import brainrender
from loguru import logger
from brainrender.camera import get_camera_params


class CameraControl:
    """
    Collection of functions to implement the actions of buttons
    aimed at controlling the camera of the brainrender Scene
    in the GUI.
    """

    def __init__(self):
        return

    def _update_camera(self):
        if self.camera_orientation is not None:
            camera = self.camera_orientation
            self.camera_orientation = None
        else:
            camera = get_camera_params(scene=self.scene)

        self.scene.render(camera=camera)
        self._update()

        logger.debug(f"GUI: resetting camera to: {self.camera_orientation}")

    def reset_camera(self):
        self.camera_orientation = brainrender.settings.DEFAULT_CAMERA
        self._update_camera()

    def move_camera_front(self):
        self.camera_orientation = "frontal"  # specify brainrender camera
        self._update_camera()

    def move_camera_top(self):
        self.camera_orientation = "top"  # specify brainrender camera
        self._update_camera()

    def move_camera_side1(self):
        self.camera_orientation = "sagittal"  # specify brainrender camera
        self._update_camera()

    def move_camera_side2(self):
        self.camera_orientation = "sagittal2"  # specify brainrender camera
        self._update_camera()



from qtpy.QtWidgets import QDialog, QLineEdit, QPushButton, QLabel, QVBoxLayout
from brainrender.gui.style import style, update_css


class AddRegionsWindow(QDialog):
    left = 250
    top = 250
    width = 400
    height = 300

    label_msg = (
        "Write the acronyms of brainregions  "
        + "you wish to add.\n[as 'space' separated strings (e.g.: STN TH)]"
    )

    def __init__(self, main_window, palette):
        """
        Creates a new window for user to input
        which regions to add to scene.

        Arguments:
        ----------

        main_window: reference to the App's main window
        palette: main_window's palette, used to style widgets
        """
        super().__init__()
        self.setWindowTitle("Select options")
        self.ui()
        self.main_window = main_window
        self.setStyleSheet(update_css(style, palette))

    def ui(self):
        """
        Define UI's elements
        """
        self.setGeometry(self.left, self.top, self.width, self.height)

        layout = QVBoxLayout()

        # Regions
        label = QLabel(self)
        label.setObjectName("PopupLabel")
        label.setText(self.label_msg)

        self.textbox = QLineEdit(self)

        # Alpha
        alpha_label = QLabel(self)
        alpha_label.setObjectName("PopupLabel")
        alpha_label.setText("Alpha")

        self.alpha_textbox = QLineEdit(self)
        self.alpha_textbox.setText(str(1.0))

        # Color
        color_label = QLabel(self)
        color_label.setObjectName("PopupLabel")
        color_label.setText("Color")

        self.color_textbox = QLineEdit(self)
        self.color_textbox.setText("atlas")

        # Create a button in the window
        self.button = QPushButton("Add regions", self)
        self.button.clicked.connect(self.on_click)
        self.button.setObjectName("RegionsButton")

        layout.addWidget(label)
        layout.addWidget(self.textbox)

        layout.addWidget(alpha_label)
        layout.addWidget(self.alpha_textbox)

        layout.addWidget(color_label)
        layout.addWidget(self.color_textbox)

        layout.addWidget(self.button)

        self.setLayout(layout)
        self.show()

    def on_click(self):
        """
        On click or 'Enter' get the regions
        from the input and call the add_regions
        method of the main window
        """
        regions = self.textbox.text().split(" ")
        self.main_window.add_regions(
            regions, self.alpha_textbox.text(), self.color_textbox.text()
        )

        self.close()



from pathlib import Path
from rich.progress import track
from rich import print
import os
import numpy as np
from myterial import amber, orange, salmon
from loguru import logger

from brainrender.camera import check_camera_param, get_camera_params
from brainrender._video import Video
import brainrender as br
from brainrender._jupyter import not_on_jupyter


class VideoMaker:
    def __init__(
        self,
        scene,
        save_fld,
        name,
        fmt="mp4",
        size="1620x1050",
        make_frame_func=None,
    ):
        """
        Creates a video by animating a scene and saving a sequence
        of screenshots.

        :param scene: the instance of Scene to be animated
        :param save_fld: str, Path. Where the video will be savd
        :param save_name: str, name of the video
        :param fmt: str. Video format (e.g. 'mp4')
        :param make_frame_func: None, optional. If passed it should be a
            function that takes the Scene to be animated as the fist argument abd
            the current frame number as second. At every frame this function
            can do what's needed to animate the scene
        :param size: str, size of video's frames in pixels
        """
        logger.debug(
            f"Creating video with name {name}. Format: {fmt}, size: {size}, save folder: {save_fld}"
        )

        self.scene = scene

        self.save_fld = Path(save_fld)
        self.save_fld.mkdir(exist_ok=True)
        self.save_name = name
        self.video_format = fmt
        self.size = size
        if "mp4" not in self.video_format:
            raise NotImplementedError(
                "Video creation can only output mp4 videos for now"
            )

        self.make_frame_func = make_frame_func or self._make_frame

    @staticmethod
    def _make_frame(
        scene, frame_number, tot_frames, azimuth=0, elevation=0, roll=0
    ):
        """
        Default `make_frame_func`. Rotaets the camera in 3 directions

        :param scene: scene to be animated.
        :param frame_number: int, not used
        :param tot_frames: int, total numner of frames
        :param azimuth: integer, specify the rotation in degrees
                    per frame on the relative axis. (Default value = 0)
        :param elevation: integer, specify the rotation in degrees
                    per frame on the relative axis. (Default value = 0)
        :param roll: integer, specify the rotation in degrees
                    per frame on the relative axis. (Default value = 0)
        """
        scene.plotter.show(interactive=False)
        scene.plotter.camera.Elevation(elevation)
        scene.plotter.camera.Azimuth(azimuth)
        scene.plotter.camera.Roll(roll)

    def generate_frames(self, fps, duration, video, *args, **kwargs):
        """
        Loop to generate frames

        :param fps: int, frame rate
        :param duration: float, video duration in seconds
        :param video: vedo Video class used to create the video
        """
        nframes = int(fps * duration)
        for i in track(range(nframes), description="Generating frames"):
            self.make_frame_func(self.scene, i, nframes, *args, **kwargs)
            video.addFrame()

    def compress(self, temp_name):
        """
        Compresses created video with ffmpeg and removes
        uncompressed video
        """

        command = f"ffmpeg -hide_banner -loglevel panic -i {temp_name}.mp4 -vcodec libx264 -crf 28 {self.save_name}.mp4 -y"
        os.system(command)
        Path(temp_name + ".mp4").unlink()

        print(
            f"[{amber}]Saved compressed video at: [{orange} bold]{self.save_fld}/{self.save_name}.{self.video_format}"
        )

    @not_on_jupyter
    def make_video(
        self, *args, duration=10, fps=30, render_kwargs={}, **kwargs
    ):
        """
        Creates a video using user defined parameters

        :param *args: any extra argument to be bassed to `make_frame_func`
        :param duration: float, duratino of the video in seconds
        :param fps: int, frame rate
        :param **kwargs: any extra keyword argument to be bassed to `make_frame_func`
        """
        logger.debug(f"Saving a video {duration}s long ({fps} fps)")
        _off = br.settings.OFFSCREEN
        br.settings.OFFSCREEN = True  # render offscreen

        self.scene.render(interactive=False, **render_kwargs)

        # cd to folder where the video will be saved
        curdir = os.getcwd()
        os.chdir(self.save_fld)
        print(f"[{amber}]Saving video in [{orange}]{self.save_fld}")

        # Create video
        video = Video(
            name=self.save_name,
            duration=duration,
            fps=fps,
            fmt=self.video_format,
            size=self.size,
        )

        # Make frames
        self.generate_frames(fps, duration, video, *args, **kwargs)
        self.scene.close()

        # Stitch frames into uncompressed video
        out, command = video.close()
        spath = f"{self.save_fld}/{self.save_name}.{self.video_format}"
        if out:
            print(
                f"[{orange} bold]ffmpeg returned an error while trying to save video with command:\n    [{salmon}]{command}"
            )
        else:
            print(f"[{amber}]Saved video at: [{orange} bold]{spath}")

        # finish up
        br.settings.OFFSCREEN = _off
        os.chdir(curdir)
        return spath


def sigma(x):
    """
    Sigmoid curve
    """
    y = 1.05 / (1 + np.exp(-8 * (x - 0.5))) - 0.025
    if y < 0:
        y = 0
    if y > 1:
        y = 1
    return y


class Animation(VideoMaker):
    """
    The animation class facilitates the creation of videos
    by specifying a series of keyframes at given moments during
    the video. At each keyframe various parameters (e.g. camera position)
    is specified and the video is created by interpolating
    between consecutive key frames.
    """

    _last_frame_params = None
    _first_zoom = 0

    def __init__(self, scene, save_fld, name, fmt="mp4", size="1620x1050"):
        """
        The animation class facilitates the creation of videos
        by specifying a series of keyframes at given moments during
        the video. At each keyframe various parameters (e.g. camera position)
        is specified and the video is created by interpolating
        between consecutive key frames.

        :param scene: the instance of Scene to be animated
        :param save_fld: str, Path. Where the video will be savd
        :param save_name: str, name of the video
        :param fmt: str. Video format (e.g. 'mp4')
        """
        VideoMaker.__init__(self, scene, save_fld, name, fmt=fmt, size=size)
        logger.debug(f"Creating animation")

        self.keyframes = {}
        self.keyframes[0] = dict(  # make sure first frame is a keyframe
            zoom=None, camera=None, callback=None
        )

    def add_keyframe(
        self,
        time,
        duration=0,
        camera=None,
        zoom=None,
        interpol="sigma",
        callback=None,
        **kwargs,
    ):
        """
        Add a keyframe to the video.

        :param time: float, time in seconds during the video
            at which the keyframe takes place.
        :param duration: float, if >0 the key frame is repeated
            every 5ms to go from start to start+duration
        :param zoom: camera zoom
        :param camera: dictionary of camera parameters
        :param interpol: str, if `sigma` or `linear` specifies
            the interpolation mode between key frames.
        :param callback: function which takes scene, current video
            frame and total number of frames in video as arguments.
            can be used to make stuff happen during a key frame (e.g. remove
            an actor)
        """
        if camera is not None:
            camera = check_camera_param(camera)

        if time in self.keyframes.keys() and time > 0:
            print(f"[b {orange}]Keyframe {time} already exists, overwriting!")

        if zoom is None:
            previous_zoom = list(self.keyframes.values())[0]["zoom"] or 0
            zoom = previous_zoom

        if not duration:
            self.keyframes[time] = dict(
                zoom=zoom,
                camera=camera,
                callback=callback,
                interpol=interpol,
                kwargs=kwargs,
            )
        else:
            for time in np.arange(time, time + duration, 0.001):
                self.keyframes[time] = dict(
                    zoom=zoom if time == 0 else None,
                    camera=camera,
                    callback=callback,
                    interpol=interpol,
                    kwargs=kwargs,
                )

    def get_keyframe_framenumber(self, fps):
        """
        Keyframes are defines in units of time (s), so we need
        to know to which frame each keyframe corresponds

        :param fps: int, frame rate
        """
        self.keyframes = {
            int(np.floor(s * fps)): v for s, v in self.keyframes.items()
        }
        self.keyframes_numbers = sorted(list(self.keyframes.keys()))

    def generate_frames(self, fps, duration, video):
        """
        Loop to generate frames

        :param fps: int, frame rate
        :param duration: float, video duration in seconds
        :param video: vedo Video class used to create the video
        """
        logger.debug(
            f"Generating animation keyframes. Duration: {duration}, fps: {fps}"
        )
        self.get_keyframe_framenumber(fps)

        self.nframes = int(fps * duration)
        self.last_keyframe = max(self.keyframes_numbers)

        if self.last_keyframe > self.nframes:
            print(
                f"[b {orange}]The video will be {self.nframes} frames long, but you have defined keyframes after that, try increasing video duration?"
            )

        for framen in track(
            range(self.nframes), description="Generating frames..."
        ):
            self._make_frame(framen)

            if framen > 1:
                video.addFrame()

    def get_frame_params(self, frame_number):
        """
        Get current parameters (e.g. camera position)
        based on frame numbe and defined key frames.

        If frame number is a keyframe or is after a keyframe
        then the params are those of that/the last keyframe.
        Else the params of two consecutive keyframes are interpolate
        using either a linear or sigmoid function.
        """
        if frame_number in self.keyframes_numbers:
            # Check if current frame is a key frame
            params = self.keyframes[frame_number]

        elif frame_number > self.last_keyframe:
            # check if current frame is past the last keyframe
            params = self.keyframes[self.last_keyframe]
            params["callback"] = None

        else:
            # interpolate between two key frames
            prev = [n for n in self.keyframes_numbers if n < frame_number][-1]
            nxt = [n for n in self.keyframes_numbers if n > frame_number][0]
            kf1, kf2 = self.keyframes[prev], self.keyframes[nxt]

            self.segment_fact = (nxt - frame_number) / (nxt - prev)
            if kf2["interpol"] == "sigma":
                self.segment_fact = sigma(self.segment_fact)

            params = dict(
                camera=self._interpolate_cameras(kf1["camera"], kf2["camera"]),
                zoom=self._interpolate_values(kf1["zoom"], kf2["zoom"]),
                callback=None,
            )

        # get current camera (to avoid using scene's default)
        if params["camera"] is None:
            params["camera"] = get_camera_params(self.scene)
        return params

    def _make_frame(self, frame_number):
        """
        Creates a frame with the correct params
        and calls the keyframe callback function if defined.

        :param frame_number: int, current frame number
        """
        frame_params = self.get_frame_params(frame_number)
        logger.debug(f"Frame {frame_number}, params: {frame_params}")

        # callback
        if frame_params["callback"] is not None:
            callback_camera = frame_params["callback"](
                self.scene,
                frame_number,
                self.nframes,
                **frame_params["kwargs"],
            )
        else:
            callback_camera = None

        # see if callback returned a camera
        camera = callback_camera or frame_params["camera"]

        # render
        self.scene.render(
            camera=camera.copy(),
            zoom=frame_params["zoom"],
            interactive=False,
            resetcam=False,
        )

    def _interpolate_cameras(self, cam1, cam2):
        """
        Interpolate the parameters of two cameras
        """
        if cam1 is None:
            return cam2
        elif cam2 is None:
            return cam1

        interpolated = {}
        for k, v1 in cam1.items():
            try:
                interpolated[k] = self._interpolate_values(v1, cam2[k])
            except KeyError:  # pragma: no cover
                raise ValueError(
                    "Cameras to interpolate dont have the same set of parameters"
                )
        return interpolated

    def _interpolate_values(self, v1, v2):
        """
        Interpolate two valuess
        """
        if v1 is None:
            return v2
        elif v2 is None:
            return v1

        return self.segment_fact * np.array(v1) + (
            1 - self.segment_fact
        ) * np.array(v2)



from bg_atlasapi.bg_atlas import BrainGlobeAtlas
from vedo import Plane
import numpy as np
from loguru import logger

from brainrender import settings
from brainrender.actor import Actor
from brainrender._io import load_mesh_from_file
from brainrender._utils import return_list_smart


class Atlas(BrainGlobeAtlas):
    def __init__(self, atlas_name=None):
        """
        Brainrender's Atlas class subclasses BrainGlobeAtlas
        to add methods to get regions meshes as Actors
        and to get a plane at a given point and normal.

        :param atlas_name: str, atlas name from brainglobe's atlas API atlases
        """
        atlas_name = atlas_name or settings.DEFAULT_ATLAS
        self.atlas_name = atlas_name
        logger.debug(f"Generating ATLAS: {atlas_name}")

        BrainGlobeAtlas.__init__(
            self, atlas_name=atlas_name, print_authors=False
        )

    @property
    def zoom(self):
        """
        Returns the best camera zoom given the atlas resolution
        """
        res = np.max(self.metadata["resolution"])

        if self.atlas_name == "allen_human_500um":
            logger.debug(
                "ATLAS: setting zoom manually for human atlas, atlas needs fixing"
            )
            return 350
        else:
            return 40 / res

    def _get_region_color(self, region):
        """
        Get's the rgb color of a region in the atlas
        """
        return [
            x / 255 for x in self._get_from_structure(region, "rgb_triplet")
        ]

    def get_region(self, *regions, alpha=1, color=None):
        """
        Get brain regions meshes as Actors
        :param regions: str with names of brain regions in the atlas
        :param alpha: float
        :param color: str
        """
        if not regions:
            return None

        _color = color
        actors = []
        for region in regions:
            if (
                region not in self.lookup_df.acronym.values
                and region not in self.lookup_df["id"].values
            ):
                print(
                    f"The region {region} doesn't seem to belong to the atlas being used: {self.atlas_name}. Skipping"
                )
                continue

            # Get mesh
            obj_file = str(self.meshfile_from_structure(region))
            mesh = load_mesh_from_file(obj_file, color=color, alpha=alpha)

            # Get color
            color = color or self._get_region_color(region)

            # Make actor
            actor = Actor(mesh, name=region, br_class="brain region")
            actor.c(color).alpha(alpha)
            actors.append(actor)

            # reset color to input
            color = _color

        return return_list_smart(actors)

    def get_plane(
        self,
        pos=None,
        norm=None,
        plane=None,
        sx=None,
        sy=None,
        color="lightgray",
        alpha=0.25,
        **kwargs,
    ):
        """
        Returns a plane going through a point at pos, oriented
        orthogonally to the vector norm and of width and height
        sx, sy.

        :param pos: 3-tuple or list with x,y,z, coords of point the plane goes through
        :param norm: 3-tuple with plane's normal vector (optional)
        :param sx, sy: int, width and height of the plane
        :param plane: "sagittal", "horizontal", or "frontal"
        :param color, alpha: plane color and transparency
        """
        axes_pairs = dict(sagittal=(0, 1), horizontal=(2, 0), frontal=(2, 1))

        if pos is None:
            pos = self.root._mesh.centerOfMass()

        try:
            norm = norm or self.space.plane_normals[plane]
        except KeyError:  # pragma: no cover
            raise ValueError(  # pragma: no cover
                f"Could not find normals for plane {plane}. Atlas space provides these normals: {self.space.plane_normals}"  # pragma: no cover
            )

        # Get plane width and height
        idx_pair = (
            axes_pairs[plane]
            if plane is not None
            else axes_pairs["horizontal"]
        )

        bounds = self.root.bounds()
        root_bounds = [
            [bounds[0], bounds[1]],
            [bounds[2], bounds[3]],
            [bounds[4], bounds[5]],
        ]

        wh = [float(np.diff(root_bounds[i])) for i in idx_pair]
        if sx is None:
            sx = wh[0]
        if sy is None:
            sy = wh[1]

        # return plane
        return Actor(
            Plane(pos=pos, normal=norm, sx=sx, sy=sy, c=color, alpha=alpha),
            name=f"Plane at {pos} norm: {norm}",
            br_class="plane",
        )



import pandas as pd
from vedo.shapes import Tube, Spheres
from vedo import merge
from loguru import logger
import numpy as np
from pathlib import Path


from brainrender.actor import Actor


def make_streamlines(
    *streamlines, color="salmon", alpha=1, radius=10, show_injection=True
):
    """
    Creates instances of Streamlines from data.
    :param streamlines: pd.dataframes with streamlines data
    :param radius: float. Radius of the Tube mesh used to render streamlines
    :param color: str, name of the color to be used
    :param alpha: float, transparancy
    :param show_injection: bool. If true spheres mark the injection sites
    """
    return [
        Streamlines(
            s,
            color=color,
            alpha=alpha,
            radius=radius,
            show_injection=show_injection,
        )
        for s in streamlines
    ]


class Streamlines(Actor):
    """
    Streamliens actor class.
    Creates an actor from streamlines data (from a json file parsed with: get_streamlines_data)
    """

    def __init__(
        self,
        data,
        radius=10,
        color="salmon",
        alpha=1,
        show_injection=True,
        name=None,
    ):
        """
        Turns streamlines data to a mesh.
        :param data: pd.DataFrame with streamlines points data
        :param radius: float. Radius of the Tube mesh used to render streamlines
        :param color: str, name of the color to be used
        :param alpha: float, transparancy
        :param name: str, name of the actor.
        :param show_injection: bool. If true spheres mark the injection sites
        """
        logger.debug(f"Creating a streamlines actor")
        if isinstance(data, (str, Path)):
            data = pd.read_json(data)
        elif not isinstance(data, pd.DataFrame):
            raise TypeError("Input data should be a dataframe")

        self.radius = radius
        mesh = (
            self._make_mesh(data, show_injection=show_injection)
            .c(color)
            .alpha(alpha)
            .clean()
        )

        name = name or "Streamlines"
        Actor.__init__(self, mesh, name=name, br_class="Streamliness")

    def _make_mesh(self, data, show_injection=True):
        lines = []
        if len(data["lines"]) == 1:
            try:
                lines_data = data["lines"][0]
            except KeyError:  # pragma: no cover
                lines_data = data["lines"]["0"]  # pragma: no cover
        else:
            lines_data = data["lines"]

        for line in lines_data:
            points = [[lin["x"], lin["y"], lin["z"]] for lin in line]
            lines.append(
                Tube(
                    points,
                    r=self.radius,
                    res=8,
                )
            )

        if show_injection:
            coords = np.vstack(
                [
                    list(point.values())
                    for point in data.injection_sites.iloc[0]
                ]
            )
            lines.append(
                Spheres(
                    coords,
                    r=self.radius * 10,
                    res=8,
                )
            )

        return merge(*lines)



from loguru import logger

import brainrender
from brainrender.gui.widgets.add_regions import AddRegionsWindow
from brainrender.gui.widgets.actors_list import remove_from_list
from brainrender.gui.utils import (
    get_region_actors,
    get_color_from_string,
    get_alpha_from_string,
)


class RegionsControl:
    def __init__(self):
        """
        Collections of functions to control the
        addition of regions meshes to the brainrender
        Scene for the GUI
        """
        return

    def open_regions_dialog(self):
        """
        Opens a QDialog window for inputting
        regions to add to the scene
        """
        self.regions_dialog = AddRegionsWindow(self, self.palette)

    def add_regions(self, regions, alpha, color):
        """
        Called by AddRegionsWindow when it closes.
        It adds brain regions to the scene

        Arguments:
        ----------
        regions: list of strings with regions acronyms
        alpha: str, meshes transparency
        color: str, meshes color. If 'atlas' the default colors are used
        """
        logger.debug(f"GUI: Adding brain regions: {regions}")
        # Get params
        alpha = get_alpha_from_string(alpha)
        if alpha is None:
            alpha = brainrender.DEFAULT_MESH_ALPHA

        color = get_color_from_string(color)
        if color == "atlas":
            colors = None
        else:
            colors = color

        # Add brain regions
        self.scene.add_brain_region(
            *regions,
            alpha=alpha,
            color=colors,
        )

        # update
        self._update()

    def add_region_from_tree(self, val):
        """
        When an item on the hierarchy tree is double clicked, the
        corresponding mesh is added/removed from the brainrender scene
        """
        logger.debug("GUI: adding brain region from tree view")
        # Get item
        idxs = self.treeView.selectedIndexes()
        if idxs:
            item = idxs[0]
        else:
            return
        item = item.model().itemFromIndex(val)

        # Get region name
        region = str(item.tag).split(" ")[0]

        # Toggle checkbox
        if not item._checked:
            # item.setCheckState(Qt.Checked)
            item._checked = True
        else:
            # item.setCheckState(Qt.Unchecked)
            item._checked = False

        # Add/remove mesh
        if get_region_actors(self.scene.actors, region) is None:
            # Add region
            self.scene.add_brain_region(
                region,
            )
        else:
            # remove regiona and update list
            # del get_region_actors(self.scene.actors, region)
            del self.actors[region]
            remove_from_list(self.actors_list, region)
            del self.scene.actors[[actor.name for actor in self.scene.actors].index(region)]


        # Update hierarchy's item font
        item.toggle_active()

        # Update brainrender scene
        self._update()



import numpy as np


def get_region_actors(actors, region):
    acts = [act for act in actors if act.name == region]

    if acts:
        return acts[0]
    else:
        return None


def get_color_from_string(string):
    try:
        rgb = string.replace("[", "").replace("]", "")
        color = np.array([float(c) for c in rgb.split(" ")])
        return color
    except ValueError:
        return string.lower()


def get_alpha_from_string(string):
    try:
        return float(string)
    except ValueError:
        return None



from qtpy.QtWidgets import QFileDialog
from pathlib import Path
import numpy as np
from loguru import logger

from brainrender.gui.widgets.add_from_file import AddFromFileWindow
from brainrender.gui.utils import (
    get_color_from_string,
    get_alpha_from_string,
)
from brainrender.actors import Points


class AddFromFile:
    def __init__(self):
        """
        Collection of functions to load data from files
        and add it to the GUI's brainrender Scene.
        """
        return

    def __add_from_file(self, fun, name_from_file=True):
        """
        General function for selecting, loading
        and adding to scene a file.

        Arguments:
        -----------

        fun: function. One of Scene's methods used to add the file's
                content to the scene.
        name_from_file: bool, optional. If True the actor's name is the name of the files loaded
        """
        options = QFileDialog.Options()
        # options |= QFileDialog.DontUseNativeDialog

        fname, _ = QFileDialog.getOpenFileName(
            self,
            "QFileDialog.getOpenFileName()",
            "",
            "All Files (*)",
            options=options,
        )

        if not fname:
            return
        else:
            # Get actor color and alpha
            dialog = AddFromFileWindow(self, self.palette)
            dialog.exec()

            alpha = get_alpha_from_string(dialog.alpha_textbox.text())
            color = get_color_from_string(dialog.color_textbox.text())

            # Add actor
            act = fun(fname)
            if not isinstance(act, (tuple, list)):
                act = [act]

            # Edit actor
            for actor in act:
                actor.name = Path(fname).name

                if color != "default":
                    actor.mesh.c(color)

                if alpha is not None:
                    actor.mesh.alpha(alpha)

            # Update
            self._update()

    def add_from_file_object(self):
        """
        Add to scene from brainrender.stl, .obj and .vtk files.
        Method of the corresponding button
        """
        logger.debug("GUI: adding mesh (e.g. obj) form file")
        self.__add_from_file(self.scene.add)

    def _get_cells_mesh(self, fp):
        data = np.load(fp)
        return self.scene.add(Points(data))

    def add_from_file_cells(self):
        """
        Add to scene from brainrender.npy files with cell coordinates data.
        Method of the corresponding button
        """
        logger.debug("GUI: adding CELLS from file")
        self.__add_from_file(self._get_cells_mesh)



import numpy as np
from qtpy.QtGui import QColor, QIcon
from pkg_resources import resource_filename
from loguru import logger

from brainrender.gui.utils import (
    get_color_from_string,
    get_alpha_from_string,
)
from brainrender.gui.style import palette


class ActorsControl:
    def __init__(self):
        """
        Collection of functions to control actors properties
        and related widget in the GUI
        """
        return

    def update_actor_properties(self):
        """
        Called when the text boxes for showing/editing
        the selected actor's alpha/color are edited.
        This function checks that the values makes sense
        and update the atuple of the selected actor.
        """
        # Get currently selected actor
        aname = self.actors_list.currentItem().text()
        logger.debug(f"Updating properties of actor: {aname}")
        if aname not in self.actors.keys():
            raise ValueError(f"Actor {aname} not in the actors record")
        else:
            actor = self.actors[aname]

        # Get color
        if not self.color_textbox.text():
            return
        color = get_color_from_string(self.color_textbox.text())

        # Get alpha
        alpha = get_alpha_from_string(self.alpha_textbox.text())
        if alpha is None:
            return

        # Update actor
        try:
            self.actors[aname] = self.atuple(
                actor.mesh, actor.is_visible, color, alpha
            )
            self._update()
        except IndexError:  # something went wrong with getting of color
            self.actors[aname] = actor
            return

    def actor_list_double_clicked(self, listitem):
        """
        When an item in the actors list is doube clicked
        it toggles the corresponding actor's visibility
        and updates the list widget UI
        """
        # Get actor
        aname = self.actors_list.currentItem().text()
        logger.debug(f"GUI: toggling {aname} visibility")
        if aname not in self.actors.keys():
            raise ValueError(f"Actor {aname} not in the actors record")
        else:
            actor = self.actors[aname]

        # Toggle visibility
        self.actors[aname] = self.atuple(
            actor.mesh, not actor.is_visible, actor.color, actor.alpha
        )

        # Toggle list item UI
        if self.actors[aname].is_visible:
            txt = palette["text"]
            icon = QIcon(resource_filename("brainrender.gui", "icons/eye.svg"))
        else:
            txt = palette["primary"]
            icon = QIcon(
                resource_filename("brainrender.gui", "icons/eye-slash.svg")
            )
        rgb = txt.replace(")", "").replace(" ", "").split("(")[-1].split(",")

        listitem.setForeground(QColor(*[int(r) for r in rgb]))
        listitem.setIcon(icon)

        # update
        self._update()

    def actor_list_clicked(self, index):
        """
        When an item of the actors list is clicked
        this function loads it's parameters and updates
        the text in the alpha/color textboxes.
        """
        # Get actor
        aname = self.actors_list.currentItem().text()
        if aname not in self.actors.keys():
            raise ValueError(f"Actor {aname} not in the actors record")
        else:
            actor = self.actors[aname]

        self.alpha_textbox.setText(str(actor.alpha))

        if isinstance(actor.color, np.ndarray):
            color = "".join(
                [str(round(c, 1)) + " " for c in actor.color]
            ).rstrip()
        else:
            color = actor.color

        self.color_textbox.setText(color)



from brainrender.actors.points import Points, Point, PointsDensity
from brainrender.actors.ruler import ruler, ruler_from_surface
from brainrender.actors.neurons import Neuron, make_neurons
from brainrender.actors.cylinder import Cylinder
from brainrender.actors.volume import Volume
from brainrender.actors.streamlines import Streamlines



import vedo
from functools import update_wrapper, partial
from brainrender import settings
from rich import print
from rich.syntax import Syntax
from myterial import orange_dark, salmon


class JupyterMixIn:  # pragma: no cover
    def __init__(self):  # pragma: no cover
        # keep track if we are in a jupyter notebook
        if vedo.settings.notebookBackend == "k3d":
            self.backend = "k3d"
        elif vedo.settings.notebookBackend == "itkwidgets":
            self.backend = "itkwidgets"
        else:
            self.backend = False

        # Can't use cartoon shader in a notebook
        if self.backend == "k3d" or self.backend == "itkwidgets":
            if settings.SHADER_STYLE == "cartoon":
                settings.SHADER_STYLE = "plastic"


class not_on_jupyter:  # pragma: no cover
    def __init__(self, func):  # pragma: no cover
        """
        A decorator to block some methods from
        running in jupyter notebooks
        """
        update_wrapper(self, func)
        self.func = func

    def __get__(self, obj, objtype):  # pragma: no cover
        """Support instance methods."""
        return partial(self.__call__, obj)

    def __call__(self, obj, *args, **kwargs):  # pragma: no cover
        backend = JupyterMixIn().backend
        if not backend:
            return self.func(obj, *args, **kwargs)
        else:
            print(
                f"[{orange_dark}]Cannot run function [bold {salmon}]{self.func.__name__}[/ bold {salmon}] in a jupyter notebook",
                f"[{orange_dark}]Try setting the correct backend before creating your scene:\n",
                Syntax("from vedo import embedWindow", lexer_name="python"),
                Syntax("embedWindow(None)", lexer_name="python"),
            )
            return None



from pathlib import Path
from morphapi.morphology.morphology import Neuron as MorphoNeuron
from vedo import Mesh
from loguru import logger
from pyinspect.utils import _class_name

from brainrender.actor import Actor


def make_neurons(
    *neurons, alpha=1, color=None, neurite_radius=8, soma_radius=15, name=None
):
    """
    Returns a list of Neurons given a variable numnber of inputs
    :param neurons: any accepted data input for Neuron
    :param alpha: float
    :param color: str
    :param neuron_radius: float, radius of axon/dendrites
    :param soma_radius: float, radius of soma
    :param name: str, actor name
    """
    return [
        Neuron(
            n,
            alpha=alpha,
            color=color,
            neurite_radius=neurite_radius,
            soma_radius=soma_radius,
            name=name,
        )
        for n in neurons
    ]


class Neuron(Actor):
    def __init__(
        self,
        neuron,
        color=None,
        alpha=1,
        neurite_radius=8,
        soma_radius=15,
        name=None,
    ):
        """
        Creates an Actor representing a single neuron's morphology
        :param neuron: path to .swc file, Mesh, Actor or Neuron from morphapi.morphology
        :param alpha: float
        :param color: str,
        :param neuron_radius: float, radius of axon/dendrites
        :param soma_radius: float, radius of soma
        :param name: str, actor name
        """
        logger.debug(f"Creating a Neuron actor")
        if color is None:
            color = "blackboard"
        alpha = alpha
        self.neurite_radius = neurite_radius
        self.soma_radius = soma_radius
        self.name = None

        if isinstance(neuron, (str, Path)):
            mesh = self._from_file(neuron)
        elif isinstance(neuron, (Mesh)):
            mesh = neuron
        elif isinstance(neuron, Actor):
            mesh = neuron.mesh
        elif isinstance(neuron, MorphoNeuron):
            mesh = self._from_morphapi_neuron(neuron)
        else:
            raise ValueError(
                f'Argument "neuron" is not in a recognized format: {_class_name(neuron)}'
            )

        Actor.__init__(self, mesh, name=self.name, br_class="Neuron")
        self.mesh.c(color).alpha(alpha)

    def _from_morphapi_neuron(self, neuron: (MorphoNeuron)):
        mesh = neuron.create_mesh(
            neurite_radius=self.neurite_radius,
            soma_radius=self.soma_radius,
        )[1]
        return mesh

    def _from_file(self, neuron: (str, Path)):
        path = Path(neuron)
        if not path.exists():
            raise FileExistsError(f"Neuron file doesnt exist: {path}")

        if not path.suffix == ".swc":
            raise NotImplementedError(
                "Neuron can load morphology only from brainrender.swc files"
            )

        self.name = self.name or path.name

        return self._from_morphapi_neuron(MorphoNeuron(data_file=neuron))



from vedo import settings as vsettings
import sys

DEBUG = False  # set to True to see more detailed logs

# ------------------------------- vedo settings ------------------------------ #

vsettings.pointSmoothing = False
vsettings.lineSmoothing = False
vsettings.polygonSmoothing = False
vsettings.immediateRendering = False

vsettings.useDepthPeeling = True
vsettings.alphaBitPlanes = 1
vsettings.maxNumberOfPeels = 12
vsettings.occlusionRatio = 0.1
vsettings.multiSamples = 0 if sys.platform == "darwin" else 8

vsettings.useSSAO = True

# For transparent background with screenshots
vsettings.screenshotTransparentBackground = False  # vedo for transparent bg
vsettings.useFXAA = True  # This needs to be false for transparent bg


# --------------------------- brainrender settings --------------------------- #

BACKGROUND_COLOR = "white"
DEFAULT_ATLAS = "allen_mouse_25um"  # default atlas
DEFAULT_CAMERA = "three_quarters"  # Default camera settings (orientation etc. see brainrender.camera.py)
INTERACTIVE = True  # rendering interactive ?
LW = 2  # e.g. for silhouettes
ROOT_COLOR = [0.8, 0.8, 0.8]  # color of the overall brain model's actor
ROOT_ALPHA = 0.2  # transparency of the overall brain model's actor'
SCREENSHOT_SCALE = 1
SHADER_STYLE = "cartoon"  # affects the look of rendered brain regions: [metallic, plastic, shiny, glossy]
SHOW_AXES = True
WHOLE_SCREEN = True  # If true render window is full screen
OFFSCREEN = False



from setuptools import setup, find_namespace_packages
from os import path
from pathlib import Path
import codecs

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def read(path):
    with codecs.open(path, "r") as fp:
        return fp.read()


def get_version():
    path = Path(this_directory) / "brainrender" / "__init__.py"

    for line in read(path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


requirements = [
    "numpy",
    "pandas",
    "vedo>=2021.0.5",
    "k3d==2.7.4",
    "msgpack",
    "pyyaml>=5.3",
    "morphapi>=0.1.3.0",
    "requests",
    "bg-atlasapi>=1.0.0",
    "tables",
    "pyinspect>=0.0.8",
    "qtpy",
    "myterial",
    "loguru",
    "vtk<=9.0.0",
]

setup(
    name="brainrender",
    version=get_version(),
    description="Visualisation and exploration of brain atlases and other anatomical data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    install_requires=requirements,
    extras_require={
        "lint": [
            "black",
            "flake8",
        ],
        "nb": ["jupyter", "k3d"],
        "dev": [
            "pytest-cov",
            "pytest-qt",
            "pytest",
            "pytest-sugar",
            "coveralls",
            "coverage<=4.5.4",
            "black",
            "flake8",
            "pre-commit",
            "opencv-python",
            "jupyter",
            "allensdk",
            "PySide2>=5.12.3",
            "k3d",
            "imio",
        ],
        "pyside2": ["PySide2>=5.12.3"],
        "pyqt5": ["PyQt5>=5.12.3"],
    },
    python_requires=">=3.6",
    packages=find_namespace_packages(exclude=("tests",)),
    entry_points={
        "console_scripts": [
            "brainrender-gui = brainrender.gui.__init__:clilaunch",
        ]
    },
    include_package_data=True,
    url="https://github.com/brainglobe/brainrender",
    author="Federico Claudi, Adam Tyson, Luigi Petrucco",
    zip_safe=False,
)



from brainrender.atlas_specific.allen_brain_atlas.gene_expression import (
    GeneExpressionAPI,
)
from brainrender.atlas_specific.allen_brain_atlas.streamlines import (
    get_streamlines_for_region,
)



import pandas as pd
from rich.progress import track
from rich import print
from loguru import logger
from myterial import orange

try:
    from allensdk.api.queries.mouse_connectivity_api import (
        MouseConnectivityApi,
    )

    mca = MouseConnectivityApi()
    allen_sdk_installed = True
except ModuleNotFoundError:  # pragma: no cover
    allen_sdk_installed = False  # pragma: no cover


from brainrender._utils import listify
from brainrender._io import request
from brainrender import base_dir

streamlines_folder = base_dir / "streamlines"
streamlines_folder.mkdir(exist_ok=True)


def experiments_source_search(SOI):
    """
    Returns data about experiments whose injection was in the SOI, structure of interest
    :param SOI: str, structure of interest. Acronym of structure to use as seed for teh search
    :param source:  (Default value = True)
    """

    transgenic_id = 0  # id = 0 means use only wild type
    primary_structure_only = True

    if not allen_sdk_installed:
        print(
            f"[{orange}]Allen skd package is not installed, cannot download streamlines data."
            "Please install `allensdk` with `pip install allensdk` (note: this requires python < 3.8)"
        )
        return None

    return pd.DataFrame(
        mca.experiment_source_search(
            injection_structures=listify(SOI),
            target_domain=None,
            transgenic_lines=transgenic_id,
            primary_structure_only=primary_structure_only,
        )
    )


def get_streamlines_data(eids, force_download=False):
    """
    Given a list of expeirmental IDs, it downloads the streamline data
    from the https://neuroinformatics.nl cache and saves them as
    json files.

    :param eids: list of integers with experiments IDs
    """
    data = []
    for eid in track(eids, total=len(eids), description="downloading"):
        url = "https://neuroinformatics.nl/HBP/allen-connectivity-viewer/json/streamlines_{}.json.gz".format(
            eid
        )

        jsonpath = streamlines_folder / f"{eid}.json"

        if not jsonpath.exists() or force_download:
            response = request(url)

            # Write the response content as a temporary compressed file
            temp_path = streamlines_folder / "temp.gz"
            with open(str(temp_path), "wb") as temp:
                temp.write(response.content)

            # Open in pandas and delete temp
            url_data = pd.read_json(
                str(temp_path), lines=True, compression="gzip"
            )
            temp_path.unlink()

            # save json
            url_data.to_json(str(jsonpath))

            # append to lists and return
            data.append(url_data)
        else:
            data.append(pd.read_json(str(jsonpath)))
    return data


def get_streamlines_for_region(region, force_download=False):
    """
    Using the Allen Mouse Connectivity data and corresponding API, this function finds expeirments whose injections
    were targeted to the region of interest and downloads the corresponding streamlines data. By default, experiements
    are selected for only WT mice and onl when the region was the primary injection target.

    :param region: str with region to use for research

    """
    logger.debug(f"Getting streamlines data for region: {region}")
    # Get experiments whose injections were targeted to the region
    region_experiments = experiments_source_search(region)
    if region_experiments is None or region_experiments.empty:
        logger.debug("No experiments found from allen data")
        return None

    return get_streamlines_data(
        region_experiments.id.values, force_download=force_download
    )



import numpy as np
from vedo.colors import getColor
from vedo.colors import colors as vcolors
import random
import matplotlib.cm as cm_mpl


def map_color(value, name="jet", vmin=None, vmax=None):
    """Map a real value in range [vmin, vmax] to a (r,g,b) color scale.

    :param value: scalar value to transform into a color
    :type value: float, list
    :param name: color map name (Default value = "jet")
    :type name: str, matplotlib.colors.LinearSegmentedColorMap
    :param vmin:  (Default value = None)
    :param vmax:  (Default value = None)
    :returns: return: (r,g,b) color, or a list of (r,g,b) colors.
    """
    if vmax < vmin:
        raise ValueError("vmax should be larger than vmin")

    mp = cm_mpl.get_cmap(name=name)

    value -= vmin
    value /= vmax - vmin
    if value > 0.999:
        value = 0.999
    elif value < 0:
        value = 0
    return mp(value)[0:3]


def make_palette(N, *colors):
    """Generate N colors starting from `color1` to `color2`
    by linear interpolation HSV in or RGB spaces.
    Adapted from vedo make_palette function

    :param int: N: number of output colors.
    :param colors: input colors, any number of colors with 0 < ncolors <= N is okay.
    """
    N = int(N)

    N_input_colors = len(colors)
    if not N_input_colors:
        raise ValueError("No colors where passed to make_palette")
    if N_input_colors > N:
        raise ValueError(
            "More input colors than out colors (N) where passed to make_palette"
        )

    if N_input_colors == N:
        return colors
    else:
        # Get how many colors for each pair of colors we are interpolating over
        fractions = [
            N // N_input_colors + (1 if x < N % N_input_colors else 0)
            for x in range(N_input_colors)
        ]

        # Get pairs of colors
        cs = [np.array(getColor(col)) for col in colors]
        cs += [cs[-1]]

        output = []
        for n, (c1, c2) in enumerate(zip(cs, cs[1:])):
            cols = []
            for f in np.linspace(0, 1, fractions[n], endpoint=True):
                c = c1 * (1 - f) + c2 * f
                cols.append(c)
            output.extend(cols)
        return output


def get_random_colors(n_colors=1):
    """
    :param n_colors:  (Default value = 1)
    """
    col_names = list(vcolors.keys())
    if n_colors == 1:
        return random.choice(col_names)
    else:
        return list(random.choices(col_names, k=n_colors))



from qtpy.QtWidgets import QDialog, QLabel, QVBoxLayout
from qtpy import QtCore

from brainrender.gui.style import style, update_css


class ScreenshotModal(QDialog):
    left = 250
    top = 250
    width = 400
    height = 120

    def __init__(self, main_window, palette):
        """
        Creates a new window for user to input
        which regions to add to scene.

        Arguments:
        ----------

        main_window: reference to the App's main window
        palette: main_window's palette, used to style widgets
        """
        super().__init__()
        self.setWindowTitle("Add brain regions")
        self.ui()
        self.main_window = main_window
        self.setStyleSheet(update_css(style, palette))

        # Start timer to autoclose
        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(1500)
        self.timer.timeout.connect(self.close)
        self.timer.start()

    def ui(self):
        """
        Define UI's elements
        """
        self.setGeometry(self.left, self.top, self.width, self.height)

        layout = QVBoxLayout()

        label = QLabel(self)

        label.setStyleSheet("font-size: 18pt; font-weight: 700;")

        label.setObjectName("PopupLabel")
        label.setText("Screenshot saved")
        layout.addWidget(label)

        self.setLayout(layout)
        self.setModal(True)
        self.show()
